from math import sqrt


def symbol_times(context, size):
    r = sqrt(size - min(size / 6, 1.7)) * 0.6189
    context.move_to(-r, -r)
    context.line_to(r, r)
    context.move_to(-r, r)
    context.line_to(r, -r)



from __future__ import annotations

from collections.abc import Callable
from math import cos, pi, sin
from typing import Any

from ..array import argpass
from ..path import Path
from .chord import ChordItem, ChordValue

HALF_PI = pi * 0.5
EPSILON = 1e-6


def constant(x: Any) -> Callable[..., Any]:
    def f(*args: Any) -> Any:
        return x

    return f


def default_source(d: ChordItem) -> float:
    return d.source


def default_target(d: ChordItem) -> float:
    return d.target


def default_radius(d: dict[str, float]) -> float:
    return d["radius"]


def default_start_angle(d: ChordValue) -> float:
    return d.start_angle


def default_end_angle(d: ChordValue) -> float:
    return d.end_angle


def default_pad_angle(d: ChordValue) -> float:
    return 0


def default_arrow_head_radius(d: ChordValue) -> float:
    return 10


class Ribbon:
    """
    Ribbon generator

    Parameters
    ----------
    head_radius : Callable[..., float] | None
        Arrow head radius accessor function
    """

    def __init__(self, head_radius: Callable[..., float] | None = None):
        self._head_radius = argpass(head_radius) if callable(head_radius) else None
        self._source = argpass(default_source)
        self._target = argpass(default_target)
        self._source_radius = argpass(default_radius)
        self._target_radius = argpass(default_radius)
        self._start_angle = argpass(default_start_angle)
        self._end_angle = argpass(default_end_angle)
        self._pad_angle = argpass(default_pad_angle)
        self._context = None

    def __call__(self, *args: Any) -> str | None:
        """
        Generates a ribbon for the given arguments. The arguments are
        arbitrary; they are propagated to the ribbon generator's accessor
        functions along with the this object.

        Parameters
        ----------
        *args : Any
            Additional arguments passed through ribbon generator's accessor
            methods

        Returns
        -------
        str | None
            Generated ribbon path
        """
        s = self._source(*args)
        t = self._target(*args)
        ap = self._pad_angle(*args) * 0.5
        argv = list(args).copy() if len(args) > 0 else [None]
        argv[0] = s
        sr = self._source_radius(*argv)
        sa0 = self._start_angle(*argv) - HALF_PI
        sa1 = self._end_angle(*argv) - HALF_PI
        argv[0] = t
        tr = self._target_radius(*argv)
        ta0 = self._start_angle(*argv) - HALF_PI
        ta1 = self._end_angle(*argv) - HALF_PI

        if self._context is None:
            self._context = buffer = Path()

        if ap > EPSILON:
            if abs(sa1 - sa0) > ap * 2 + EPSILON:
                if sa1 > sa0:
                    sa0 += ap
                    sa1 -= ap
                else:
                    sa0 -= ap
                    sa1 += ap
            else:
                sa0 = sa1 = (sa0 + sa1) / 2
            if abs(ta1 - ta0) > ap * 2 + EPSILON:
                if ta1 > ta0:
                    ta0 += ap
                    ta1 -= ap
                else:
                    ta0 -= ap
                    ta1 += ap
            else:
                ta0 = ta1 = (ta0 + ta1) / 2

        self._context.move_to(sr * cos(sa0), sr * sin(sa0))
        self._context.arc(0, 0, sr, sa0, sa1)
        if sa0 != ta0 or sa1 != ta1:
            if self._head_radius is None:
                self._context.quadratic_curve_to(0, 0, tr * cos(ta0), tr * sin(ta0))
                self._context.arc(0, 0, tr, ta0, ta1)
            else:
                hr = self.head_radius(*args)
                tr2 = tr - hr
                ta2 = (ta0 + ta1) / 2
                self._context.quadratic_curve_to(0, 0, tr2 * cos(ta0), tr2 * sin(ta0))
                self._context.line_to(tr * cos(ta2), tr * sin(ta2))
                self._context.line_to(tr2 * cos(ta1), tr2 * sin(ta1))
        self._context.quadratic_curve_to(0, 0, sr * cos(sa0), sr * sin(sa0))
        self._context.close_path()

        if buffer:
            self._context = None
            return str(buffer) or None

    def set_head_radius(
        self, head_radius: Callable[..., float] | float | None = None
    ) -> Ribbon:
        """
        Sets the arrowhead radius accessor to the specified function and
        returns this ribbon generator.

        Parameters
        ----------
        head_radius : Callable[..., float] | float | None
            Arrow head radius accessor function

        Returns
        -------
        Ribbon
            Itself
        """
        if head_radius is None:
            self._head_radius = None
            return self
        if callable(head_radius):
            self._head_radius = head_radius
        else:
            self._head_radius = constant(head_radius)
        self._head_radius = argpass(self._head_radius)
        return self

    def set_radius(self, radius: Callable[..., float] | float) -> Ribbon:
        """
        Sets the source and target radius accessor to the specified function
        and returns this ribbon generator.

        Parameters
        ----------
        radius : Callable[..., float] | float
            Radius function or constant value

        Returns
        -------
        Ribbon
            Itself
        """
        if callable(radius):
            self._source_radius = radius
            self._target_radius = radius
        else:
            radius = constant(radius)
            self._source_radius = radius
            self._target_radius = radius
        self._source_radius = argpass(self._source_radius)
        self._target_radius = argpass(self._target_radius)
        return self

    def set_source_radius(self, source_radius: Callable[..., float] | float) -> Ribbon:
        """
        Sets the source radius accessor to the specified function and returns
        this ribbon generator.

        Parameters
        ----------
        source_radius : Callable[..., float] | float
            Source radius function

        Returns
        -------
        Ribbon
            Itself
        """
        if callable(source_radius):
            self._source_radius = source_radius
        else:
            self._source_radius = constant(source_radius)
        self._source_radius = argpass(self._source_radius)
        return self

    def set_target_radius(self, target_radius: Callable[..., float] | float) -> Ribbon:
        """
        Sets the target radius accessor to the specified function and returns
        this ribbon generator.

        Parameters
        ----------
        target_radius : Callable[..., float] | float
            Target radius function

        Returns
        -------
        Ribbon
            Itself
        """
        if callable(target_radius):
            self._target_radius = target_radius
        else:
            self._target_radius = constant(target_radius)
        self._target_radius = argpass(self._target_radius)
        return self

    def set_start_angle(self, start_angle: Callable[..., float] | float) -> Ribbon:
        """
        Sets the start angle accessor to the specified function and returns
        this ribbon generator.

        Parameters
        ----------
        start_angle : Callable[..., float] | float
            Start angle function or constant value

        Returns
        -------
        Ribbon
            Itself
        """
        if callable(start_angle):
            self._start_angle = start_angle
        else:
            self._start_angle = constant(start_angle)
        self._start_angle = argpass(self._start_angle)
        return self

    def set_end_angle(self, end_angle: Callable[..., float] | float) -> Ribbon:
        """
        Sets the end angle accessor to the specified function and returns
        this ribbon generator.

        Parameters
        ----------
        end_angle : Callable[..., float] | float
            End angle function or constant value

        Returns
        -------
        Ribbon
            Itself
        """
        if callable(end_angle):
            self._end_angle = end_angle
        else:
            self._end_angle = constant(end_angle)
        self._end_angle = argpass(self._end_angle)
        return self

    def set_pad_angle(self, pad_angle: Callable[..., float] | float) -> Ribbon:
        """
        Sets the pad angle accessor to the specified function and returns this
        ribbon generator

        Parameters
        ----------
        pad_angle : Callable[..., float] | float
            Pad angle function or constant value

        Returns
        -------
        Ribbon
            Itself
        """
        if callable(pad_angle):
            self._pad_angle = pad_angle
        else:
            self._pad_angle = constant(pad_angle)
        self._pad_angle = argpass(self._pad_angle)
        return self

    def set_source(self, source: Callable[..., float] | float) -> Ribbon:
        """
        Sets the source accessor to the specified function and returns this
        ribbon generator.

        Parameters
        ----------
        source : Callable[..., float] | float
            Source function or constant function

        Returns
        -------
        Ribbon
            Itself
        """
        self._source = argpass(source)
        return self

    def set_target(self, target: Callable[..., float] | float) -> Ribbon:
        """
        Sets the target accessor to the specified function and returns this
        ribbon generator.

        Parameters
        ----------
        target : Callable[..., float] | float
            Target function or constant function

        Returns
        -------
        Ribbon
            Itself
        """
        self._target = argpass(target)
        return self

    def set_context(self, context: Any | None = None) -> Ribbon:
        """
        If context is specified, sets the context and returns this ribbon
        generator. If the context is not :code:`None`, then the generated
        ribbon is rendered to this context as a sequence of path method calls.
        Otherwise, a path data string representing the generated ribbon is
        returned.

        Parameters
        ----------
        context : Any | None
            Context value

        Returns
        -------
        Ribbon
            Itself
        """
        self._context = context
        return self

    def get_head_radius(self) -> Callable[..., float]:
        return self._head_radius

    def get_radius(self) -> Callable[..., float]:
        return self._source_radius

    def get_source_radius(self) -> Callable[..., float]:
        return self._source_radius

    def get_target_radius(self) -> Callable[..., float]:
        return self._target_radius

    def get_start_angle(self) -> Callable[..., float]:
        return self._start_angle

    def get_end_angle(self) -> Callable[..., float]:
        return self._end_angle

    def get_pad_angle(self) -> Callable[..., float]:
        return self._pad_angle

    def get_source(self) -> Callable[..., float]:
        return self._source

    def get_target(self) -> Callable[..., float]:
        return self._target

    def get_context(self) -> Any | None:
        return self._context


def ribbon() -> Ribbon:
    """
    Creates a new ribbon generator with the default settings.

    Returns
    -------
    Ribbon
        Ribbon generator
    """
    return Ribbon(None)


def ribbon_arrow() -> Ribbon:
    """
    Creates a new arrow ribbon generator with the default settings.
    See also :func:`d3.chord_directed <detroit.chord_directed>`.

    Returns
    -------
    Ribbon
        Ribbon generator
    """
    return Ribbon(default_arrow_head_radius)



from __future__ import annotations

from collections.abc import Callable

from .hierarchy import Node
from .treemap.dice import dice
from .treemap.round import round_node


class Partition:
    """
    Partition layout
    """

    def __init__(self):
        self._dx = 1
        self._dy = 1
        self._padding = 0
        self._round = False

    def __call__(self, root: Node) -> Node:
        """
        Lays out the specified root hierarchy, assigning the following
        properties on root and its descendants:

        * :code:`node.x0` - the left edge of the rectangle
        * :code:`node.y0` - the top edge of the rectangle
        * :code:`node.x1` - the right edge of the rectangle
        * :code:`node.y1` - the bottom edge of the rectangle

        You must call :code:`root.sum` before passing the hierarchy to the
        partition layout. You probably also want to call root.sort to order the
        hierarchy before computing the layout.

        Parameters
        ----------
        root : Node
            Root node

        Returns
        -------
        Node
            Node organized as partition
        """
        n = root.height + 1
        root.x0 = self._padding
        root.y0 = self._padding
        root.x1 = self._dx
        root.y1 = self._dy / n
        root.each_before(self._position_node(self._dy, n))
        if self._round:
            root.each_before(round_node)
        return root

    def _position_node(self, dy: float, n: int) -> Callable[[Node], None]:
        def local_function(node):
            if node.children:
                dice(
                    node,
                    node.x0,
                    dy * (node.depth + 1) / n,
                    node.x1,
                    dy * (node.depth + 2) / n,
                )
            x0 = node.x0
            y0 = node.y0
            x1 = node.x1 - self._padding
            y1 = node.y1 - self._padding
            if x1 < x0:
                x0 = x1 = (x0 + x1) * 0.5
            if y1 < y0:
                y0 = y1 = (y0 + y1) * 0.5
            node.x0 = x0
            node.y0 = y0
            node.x1 = x1
            node.y1 = y1

        return local_function

    def set_round(self, round: bool) -> Partition:
        """
        Enables or disables rounding according to the given boolean and returns
        this partition layout.

        Parameters
        ----------
        round : bool
            Round boolean value

        Returns
        -------
        Partition
            Itself
        """
        self._round = round
        return self

    def set_size(self, size: tuple[float, float]) -> Partition:
        """
        Sets this partition layout’s size to the specified two-element array of
        numbers :code:`[width, height]` and returns this partition layout.

        Parameters
        ----------
        size : tuple[float, float]
            Size values

        Returns
        -------
        Partition
            Itself
        """
        self._dx = size[0]
        self._dy = size[1]
        return self

    def set_padding(self, padding: float) -> Partition:
        """
        Sets the padding to the specified number and returns this partition
        layout.

        Parameters
        ----------
        padding : float
            Padding value

        Returns
        -------
        Partition
            Itself
        """
        self._padding = padding
        return self

    def get_round(self) -> bool:
        return self._round

    def get_size(self) -> tuple[float, float]:
        return [self._dx, self._dy]

    def get_padding(self) -> float:
        return self._padding


def partition() -> Partition:
    """
    Builds a new partition layout with the default settings.

    Returns
    -------
    Partition
        Partition object
    """
    return Partition()



from .back import (
    ease_back_in,
    ease_back_in_out,
    ease_back_out,
)
from .back import ease_back_in_out as ease_back
from .bounce import (
    ease_bounce_in,
    ease_bounce_in_out,
    ease_bounce_out,
)
from .bounce import ease_bounce_out as ease_bounce
from .circle import (
    ease_circle_in,
    ease_circle_in_out,
    ease_circle_out,
)
from .circle import ease_circle_in_out as ease_circle
from .cubic import (
    ease_cubic_in,
    ease_cubic_in_out,
    ease_cubic_out,
)
from .cubic import ease_cubic_in_out as ease_cubic
from .elastic import (
    ease_elastic_in,
    ease_elastic_in_out,
    ease_elastic_out,
)
from .elastic import ease_elastic_out as ease_elastic
from .exp import (
    ease_exp_in,
    ease_exp_in_out,
    ease_exp_out,
)
from .exp import ease_exp_in_out as ease_exp
from .linear import ease_linear
from .poly import (
    ease_poly_in,
    ease_poly_in_out,
    ease_poly_out,
)
from .poly import ease_poly_in_out as ease_poly
from .quad import (
    ease_quad_in,
    ease_quad_in_out,
    ease_quad_out,
)
from .quad import ease_quad_in_out as ease_quad
from .sin import (
    ease_sin_in,
    ease_sin_in_out,
    ease_sin_out,
)
from .sin import ease_sin_in_out as ease_sin

__all__ = [
    "ease_back",
    "ease_back_in",
    "ease_back_in_out",
    "ease_back_out",
    "ease_bounce",
    "ease_bounce_in",
    "ease_bounce_in_out",
    "ease_bounce_out",
    "ease_circle",
    "ease_circle_in",
    "ease_circle_in_out",
    "ease_circle_out",
    "ease_cubic",
    "ease_cubic_in",
    "ease_cubic_in_out",
    "ease_cubic_out",
    "ease_elastic",
    "ease_elastic_in",
    "ease_elastic_in_out",
    "ease_elastic_out",
    "ease_exp",
    "ease_exp_in",
    "ease_exp_in_out",
    "ease_exp_out",
    "ease_linear",
    "ease_poly",
    "ease_poly_in",
    "ease_poly_in_out",
    "ease_poly_out",
    "ease_quad",
    "ease_quad_in",
    "ease_quad_in_out",
    "ease_quad_out",
    "ease_sin",
    "ease_sin_in",
    "ease_sin_in_out",
    "ease_sin_out",
]



from collections.abc import Callable
from typing import Any


def optional(f: Callable[..., Any] | None) -> Callable[..., Any] | None:
    if f is None:
        return None
    return required(f)


def required(f: Callable[..., Any]) -> Callable[..., Any]:
    if not callable(f):
        raise TypeError(f"{f!r} is not callable.")
    return f



def ease_linear(t: float) -> float:
    """
    Linear easing; the identity function; :code:`d3.ease_linear(t)` returns
    :code:`t`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_linear(0.2)
    0.2
    """
    return t



from math import copysign, inf


def clip_line(
    ax: float,
    ay: float,
    bx: float,
    by: float,
    x0: float,
    y0: float,
    x1: float,
    y1: float,
) -> tuple[float, float, float, float] | None:
    t0 = 0
    t1 = 1
    dx = bx - ax
    dy = by - ay

    r = x0 - ax
    if not dx and r > 0:
        return None
    try:
        r = r / dx
    except ZeroDivisionError:
        r = copysign(r, 0) * inf
    if dx < 0:
        if r < t0:
            return None
        if r < t1:
            t1 = r
    elif dx > 0:
        if r > t1:
            return None
        if r > t0:
            t0 = r

    r = x1 - ax
    if not dx and r < 0:
        return None
    try:
        r = r / dx
    except ZeroDivisionError:
        r = copysign(r, 0) * inf
    if dx < 0:
        if r > t1:
            return None
        if r > t0:
            t0 = r
    elif dx > 0:
        if r < t0:
            return None
        if r < t1:
            t1 = r

    r = y0 - ay
    if not dy and r > 0:
        return None
    try:
        r = r / dy
    except ZeroDivisionError:
        r = copysign(r, 0) * inf
    if dy < 0:
        if r < t0:
            return None
        if r < t1:
            t1 = r
    elif dy > 0:
        if r > t1:
            return None
        if r > t0:
            t0 = r

    r = y1 - ay
    if not dy and r < 0:
        return None
    try:
        r = r / dy
    except ZeroDivisionError:
        r = copysign(r, 0) * inf
    if dy < 0:
        if r > t1:
            return None
        if r > t0:
            t0 = r
    elif dy > 0:
        if r < t0:
            return None
        if r < t1:
            t1 = r

    x1 = ax
    y1 = ay
    x2 = bx
    y2 = by
    if t0 > 0:
        x1 = ax + t0 * dx
        y1 = ay + t0 * dy
    if t1 < 1:
        x2 = ax + t1 * dx
        y2 = ay + t1 * dy
    return x1, y1, x2, y2



from ..types import Point2D


def contains(polygon: list[Point2D], point: Point2D) -> bool:
    """
    Returns :code:`True` if and only if the specified point is inside the
    specified polygon.


    Parameters
    ----------
    polygon : list[Point2D]
        Polygon
    point : Point2D
        2D Point

    Returns
    -------
    bool
        :code:`True` if and only if the specified point is inside the specified
        polygon.

    Examples
    --------
    >>> d3.polygon_contains([[1, 1], [1.5, 0], [2, 1]], [1.5, 0.667])
    True
    """
    p = polygon[-1]
    x, y = point
    x0, y0 = p
    inside = False

    for point in polygon:
        x1, y1 = point
        if (y1 > y) != (y0 > y) and x < (x0 - x1) * (y - y1) / (y0 - y1) + x1:
            inside = not inside
        x0, y0 = x1, y1

    return inside



from collections.abc import Callable
from datetime import datetime
from math import isnan

from ..types import GenValue, T


def is_null(x: float | None) -> bool:
    """
    Returns if the input value is :code:`None` or :code:`math.nan`.

    Parameters
    ----------
    x : float | None
        Input value to check

    Returns
    -------
    bool
        If value is :code:`None` or :code:`math.nan`.
    """
    return x is None or isnan(x)


def as_float(x: GenValue) -> float:
    """
    Converts the input into :code:`float` type.

    Parameters
    ----------
    x : datetime | str | int | float
        Input to be converted

    Returns
    -------
    float
        Input converted into :code:`float`
    """
    if isinstance(x, datetime):
        return x.timestamp()
    else:
        return float(x)


def constant(x: T) -> Callable[..., T]:
    """
    Takes a value :code:`x` and returns a function which takes any argument
    and gives the value :code:`x`.

    Parameters
    ----------
    x : T
        Value to be returned by the constant function

    Returns
    -------
    Callable[..., T]
        Function which takes any argument and returns the :code:`x` value
    """

    def f(*args):
        return x

    return f


def identity(x: T) -> T:
    """
    Returns the input without any modification.

    Parameters
    ----------
    x : T
        Input

    Returns
    -------
    T
        Same value as input
    """
    return x



import math
from collections.abc import Callable

from ...selection import Selection
from ...types import Number
from .cardinal_open import curve_cardinal_open
from .catmull_rom import BezierTrait
from .common import Curve, isvaluable


class CatmullRomOpenCurve(Curve, BezierTrait):
    def __init__(self, context, alpha):
        self._context = context
        self._alpha = alpha
        self._line = math.nan

    def area_start(self):
        self._line = 0

    def area_end(self):
        self._line = math.nan

    def line_start(self):
        self._x0 = math.nan
        self._y0 = math.nan
        self._x1 = math.nan
        self._y1 = math.nan
        self._x2 = math.nan
        self._y2 = math.nan
        self._l01_a = 0
        self._l12_a = 0
        self._l23_a = 0
        self._l01_2a = 0
        self._l12_2a = 0
        self._l23_2a = 0
        self._point = 0

    def line_end(self):
        if isvaluable(self._line) or (self._line != 0 and self._point == 3):
            self._context.close_path()
        self._line = 1 - self._line

    def point(self, x, y):
        if self._point != 0:
            x23 = self._x2 - x
            y23 = self._y2 - y
            self._l23_2a = math.pow(x23 * x23 + y23 * y23, self._alpha)
            self._l23_a = math.sqrt(self._l23_2a)

        if self._point == 0:
            self._point = 1
        elif self._point == 1:
            self._point = 2
        elif self._point == 2:
            self._point = 3
            if isvaluable(self._line):
                self._context.line_to(self._x2, self._y2)
            else:
                self._context.move_to(self._x2, self._y2)
        elif self._point == 3:
            self._point = 4
            self._bezier_curve_to(x, y)
        else:
            self._bezier_curve_to(x, y)

        self._l01_a = self._l12_a
        self._l12_a = self._l23_a
        self._l01_2a = self._l12_2a
        self._l12_2a = self._l23_2a
        self._x0 = self._x1
        self._x1 = self._x2
        self._x2 = x
        self._y0 = self._y1
        self._y1 = self._y2
        self._y2 = y


def curve_catmull_rom_open(
    context_or_alpha: Selection | Number,
) -> Callable[[Selection], Curve] | Curve:
    """
    Produces a cubic Catmull–Rom spline using the specified control points and
    the parameter alpha, as proposed by Yuksel et al. Unlike curveCatmullRom,
    one-sided differences are not used for the first and last piece, and thus
    the curve starts at the second point and ends at the penultimate point.
    Default alpha value is :code:`0.5`.


    Parameters
    ----------
    context_or_alpha : Selection | Number
        Context or alpha value in range :math:`[0, 1]`. If alpha is zero,
        produces a uniform spline, equivalent to curveCardinal with a tension
        of zero; if alpha is one, produces a chordal spline; if alpha is 0.5,
        produces a centripetal spline. Centripetal splines are recommended to
        avoid self-intersections and overshoot.

    Returns
    -------
    Callable[[Selection], Curve] | Curve
        Curve object or function which makes a curve object with alpha value
        set

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_catmull_rom_open)
    >>> line(points)
    'M6,8C7.209,9.067,8.868,9.733,10,10C10.757,10.179,11.438,10.405,12,10C12.999,9.280,12.631,4.987,14,4C15.334,3.039,18.318,3.303,20,4C21.633,4.676,22.477,7.687,24,8C25.486,8.305,27.578,6.748,29,6C30.164,5.388,30.965,4.139,32,4C32.969,3.870,34.063,5.221,35,5'
    >>> line = d3.line().curve(d3.curve_catmull_rom_open(0.2))
    >>> line(points)
    'M6,8C7.275,9.215,8.964,9.693,10,10C10.882,10.261,11.394,10.703,12,10C12.763,9.115,12.653,4.995,14,4C15.333,3.016,18.327,3.321,20,4C21.653,4.671,22.491,7.675,24,8C25.494,8.322,27.637,6.696,29,6C30.258,5.358,30.987,4.156,32,4C32.987,3.848,34.028,5.286,35,5'

    **Result**

    .. image:: ../../figures/light_curve_catmull_rom_open.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_catmull_rom_open.svg
       :align: center
       :class: only-dark
    """
    if isinstance(context_or_alpha, (int, float)):
        alpha = context_or_alpha
        if alpha == 0.0:
            return curve_cardinal_open(0.0)

        def local_curve(context):
            return CatmullRomOpenCurve(context, alpha)

        return local_curve
    context = context_or_alpha
    return CatmullRomOpenCurve(context, 0.5)



from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic

from ..array import argpass
from ..types import T, U, V
from .constant import constant
from .offset import offset_none
from .order import order_none
from .series import Serie, Series


def stack_value(d: U, key: V) -> T:
    """
    Default stack value function for `Stack` class

    Parameters
    ----------
    d : U
        Data
    key : V
        Series key to access specific element of the data

    Returns
    -------
    T
        Element accessed in the data
    """
    return d[key]


def stack_series(key: T) -> Series:
    """
    Default function to make a stack series

    Parameters
    ----------
    key : T
        Key of the series

    Returns
    -------
    Series
        New series with the associated key
    """
    series = Series()
    series.key = key
    return series


class Stack(Generic[T]):
    """
    Builds a new stack generator with the default settings.
    """

    def __init__(self):
        self._keys = argpass(constant([]))
        self._order = order_none
        self._offset = offset_none
        self._value = argpass(stack_value)

    def __call__(self, data: list[T], *args: Any) -> list[Series]:
        """
        Generates a stack for the given array of data and returns an array
        representing each series.

        Parameters
        ----------
        data : list[T]
            List of data
        *args : Any
            Additional arguments passed to :code:`keys` method.

        Returns
        -------
        list[Series]
            List of series
        """
        args = [data] + list(args)
        sz: list[Series] = list(map(stack_series, self._keys(*args)))
        n = len(sz)
        j = -1

        for d in data:
            j += 1
            for i in range(n):
                series = sz[i]
                series.append(Serie([0, self._value(d, series.key, j, data)], d))

        oz = list(self._order(sz))
        for i in range(n):
            sz[oz[i]].index = i

        self._offset(sz, oz)
        return sz

    def set_keys(self, keys: Callable[[...], list[str]] | list[str]) -> Stack:
        """
        Sets the :code:`keys` method and returns itself.

        Parameters
        ----------
        keys : Callable[[...], str] | list[str]
            List of keys or function which returns a list of keys

        Returns
        -------
        Stack
            Itself

        Notes
        -----
        The arguments of the function are the same as the additional arguments
        of :code:`Stack.__call__` method.
        """
        if callable(keys):
            self._keys = keys
        else:
            self._keys = constant(list(keys))
        self._keys = argpass(self._keys)
        return self

    def set_value(
        self, value: Callable[[T, str, int, list[T]], float] | float
    ) -> Stack:
        """
        Sets the :code:`value` method and returns itself.

        Parameters
        ----------
        value : Callable[[T, str, int, list[T]], float] | float
            Value number or function which takes the optional argument: data,
            key, index, list of data. The function must returns a float value.

        Returns
        -------
        Stack
            Itself

        Notes
        -----
        You can pass a function with only the data and key arguments for example.
        """
        if callable(value):
            self._value = value
        else:
            self._value = constant(float(value))
        self._value = argpass(self._value)
        return self

    def set_order(
        self, order: Callable[[list[Series]], list[int]] | list[int] | None = None
    ) -> Stack:
        """
        Sets the :code:`order` method and returns itself.

        Parameters
        ----------
        order : Callable[[list[Series]], list[int]] | list[int] | None
            If none value, resets the method. List of indices or function which
            takes in argument a list of series and returns the order list.

        Returns
        -------
        Stack
            Itself
        """
        if order is None:
            self._order = order_none
        elif callable(order):
            self._order = order
        else:
            self._order = constant(list(order))
        return self

    def set_offset(
        self, offset: Callable[[list[Series], list[int]], None] | None = None
    ) -> Stack:
        """
        Sets the :code:`offset` and returns itself.

        Parameters
        ----------
        offset : Callable[[list[Series], list[int]], None] | None
            If none value, resets the method. Function which takes in arguments
            a list of series and the order list and updates lower and upper
            values int the series list.

        Returns
        -------
        Stack
            Itself
        """
        if offset is None:
            self._offset = offset_none
        else:
            self._offset = offset
        return self

    def get_keys(self) -> Callable[[...], str]:
        return self._keys

    def get_value(self) -> Callable[[T, str, int, list[T]], float]:
        return self._value

    def get_order(self) -> Callable[[list[Series]], list[int]]:
        return self._order

    def get_offset(self) -> Callable[[list[Series], list[int]], None]:
        return self._offset



from ..array import argpass
from ..types import GeoJSON, T
from .common import PolygonStream


def get(values: list[T], index: int) -> T | None:
    return values[index] if index < len(values) else None


def stream_geometry(geometry: GeoJSON, stream: PolygonStream):
    if geometry and hasattr(StreamGeometryType, geometry["type"]):
        getattr(StreamGeometryType, geometry["type"])(geometry, stream)


class StreamObjectType:
    @staticmethod
    def Feature(obj: GeoJSON, stream: PolygonStream):
        stream_geometry(obj["geometry"], stream)

    @staticmethod
    def FeatureCollection(obj: GeoJSON, stream: PolygonStream):
        features = obj["features"]
        for feature in features:
            stream_geometry(feature["geometry"], stream)


class StreamGeometryType:
    @staticmethod
    def Sphere(obj: GeoJSON, stream: PolygonStream):
        stream.sphere()

    @staticmethod
    def Point(obj: GeoJSON, stream: PolygonStream):
        obj = obj["coordinates"]
        argpass(stream.point)(get(obj, 0), get(obj, 1), get(obj, 2))

    @staticmethod
    def MultiPoint(obj: GeoJSON, stream: PolygonStream):
        coordinates = obj["coordinates"]
        for obj in coordinates:
            argpass(stream.point)(get(obj, 0), get(obj, 1), get(obj, 2))

    @staticmethod
    def LineString(obj: GeoJSON, stream: PolygonStream):
        stream_line(obj["coordinates"], stream, False)

    @staticmethod
    def MultiLineString(obj: GeoJSON, stream: PolygonStream):
        coordinates = obj["coordinates"]
        for obj in coordinates:
            stream_line(obj, stream, False)

    @staticmethod
    def Polygon(obj: GeoJSON, stream: PolygonStream):
        stream_polygon(obj["coordinates"], stream)

    @staticmethod
    def MultiPolygon(obj: GeoJSON, stream: PolygonStream):
        coordinates = obj["coordinates"]
        for obj in coordinates:
            stream_polygon(obj, stream)

    @staticmethod
    def GeometryCollection(obj: GeoJSON, stream: PolygonStream):
        geometries = obj["geometries"]
        for obj in geometries:
            stream_geometry(obj, stream)


def stream_line(coordinates: list[T], stream: PolygonStream, closed: bool):
    stream.line_start()
    coordinates = coordinates[:-1] if closed else coordinates
    for coordinate in coordinates:
        argpass(stream.point)(
            get(coordinate, 0), get(coordinate, 1), get(coordinate, 2)
        )
    stream.line_end()


def stream_polygon(coordinates: list[T], stream: PolygonStream):
    stream.polygon_start()
    for coordinate in coordinates:
        stream_line(coordinate, stream, True)
    stream.polygon_end()


def geo_stream(obj: GeoJSON, stream: PolygonStream):
    """
    Streams the specified GeoJSON object to the specified projection stream.
    While both features and geometry objects are supported as input, the stream
    interface only describes the geometry, and thus additional feature
    properties are not visible to streams.

    Parameters
    ----------
    obj : GeoJSON
        GeoJSON object
    stream : PolygonStream
        Stream object
    """
    if obj and hasattr(StreamObjectType, obj["type"]):
        getattr(StreamObjectType, obj["type"])(obj, stream)
    else:
        stream_geometry(obj, stream)



from ...types import T
from ..series import Series


def order_reverse(series: list[Series[T]]) -> list[int]:
    """
    Returns the reverse of the given series order :math:`[n - 1, n - 2, ...,
    0]` where :math:`n` is the number of elements in series. Thus, the stack
    order is given by the reverse of the key accessor.


    Parameters
    ----------
    series : list[Series[T]]
        List of series

    Returns
    -------
    list[int]
        Order list
    """
    return list(reversed(range(len(series))))



from lxml import etree

from ..array import argpass
from ..types import Accessor, EtreeFunction, T


def str_to_attrs(style: str) -> dict[str, str]:
    """
    Transforms a string into a dictionary containing style attributes

    Parameters
    ----------
    style : str
        Style string

    Returns
    -------
    dict[str, str]
        Dictionary containing style attributes
    """
    if not style:
        return {}
    style = style[:-1] if style.endswith(";") else style
    return {
        property: value
        for property, value in (desc.split(":") for desc in style.split(";"))
    }


def attrs_to_str(attrs: dict[str, str]) -> str:
    """
    Transforms a dictionary containing style attributes into a string

    Parameters
    ----------
    attrs : dict[str, str]
        Style attributes

    Returns
    -------
    str
        Style string
    """
    style = ";".join((":".join(pair) for pair in attrs.items()))
    return f"{style};"


def style_constant(name: str, value: str) -> EtreeFunction[T, None]:
    """
    Returns a function which adds a style attribute to nodes given a constant
    value.

    Parameters
    ----------
    name : str
        Style name to add on nodes
    value : str
        Value of the style

    Returns
    -------
    EtreeFunction[T, None]
        Function which adds a style attribute to nodes
    """

    def callback(node: etree.Element, data: T, i: int, group: list[etree.Element]):
        style = node.get("style", "")
        attrs = str_to_attrs(style)
        if name in attrs:
            attrs[name] = str(value)
            style = attrs_to_str(attrs)
        else:
            style = f"{style}{name}:{value};"
        node.set("style", style)

    return callback


def style_function(name: str, value: Accessor[T, str]) -> EtreeFunction[T, None]:
    """
    Returns a function which adds a style attribute to nodes based on an
    accessor function.

    Parameters
    ----------
    name : str
        Style name to add on nodes
    value : Accessor[T, str]
        Accessor function

    Returns
    -------
    EtreeFunction[T, None]
        Function which adds a style attribute to nodes
    """
    value = argpass(value)

    def callback(node: etree.Element, data: T, i: int, group: list[etree.Element]):
        style = node.get("style", "")
        new_value = value(data, i, group)
        attrs = str_to_attrs(style)
        if name in attrs:
            attrs[name] = str(new_value)
            style = attrs_to_str(attrs)
        else:
            style = f"{style}{name}:{new_value};"
        node.set("style", style)

    return callback


def style_value(style: str, name: str) -> str:
    """
    Gets the style value in a style string value.

    Parameters
    ----------
    style : str
        Style string value
    name : str
        Name of the style

    Returns
    -------
    str
        Value of the style
    """
    if style.endswith(";"):
        style = style[:-1]
    attrs = {
        property: value
        for property, value in (desc.split(":") for desc in style.split(";"))
    }
    return attrs.get(name)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "fee6cefdae6be6550d",
    "feeddefdbe85fd8d3cd94701",
    "feeddefdbe85fd8d3ce6550da63603",
    "feeddefdd0a2fdae6bfd8d3ce6550da63603",
    "feeddefdd0a2fdae6bfd8d3cf16913d948018c2d04",
    "fff5ebfee6cefdd0a2fdae6bfd8d3cf16913d948018c2d04",
    "fff5ebfee6cefdd0a2fdae6bfd8d3cf16913d94801a636037f2704",
]
SCHEME_ORANGES = list(map(colors, scheme))
interpolate_oranges = ramp(SCHEME_ORANGES)



from .exponent import exponent


def precision_round(step: int | float, max_val: int | float) -> int:
    """
    Returns a suggested decimal precision for format types that
    round to significant digits given the specified numeric step
    and max values. The step represents the minimum absolute difference
    between values that will be formatted, and the max represents the
    largest absolute value that will be formatted.

    Parameters
    ----------
    step : int | float
        Specific numeric value
    max_val : int | float
        The largest absolute value that will be formatted

    Returns
    -------
    int
        Decimal precision

    Examples
    --------

    >>> p = d3.precision_round(0.01, 1.01)
    >>> f = d3.format(f".{p}r")
    >>> f(0.99)
    '0.990'
    >>> f(1.0)
    '1.00'
    >>> f(1.01)
    '1.01'
    """
    step = abs(step)
    max_val = abs(max_val) - step
    return max(0, exponent(max_val) - exponent(step)) + 1



from math import atan, cos, sin

from ...types import Point2D
from ..common import Projection
from .azimuthal import azimuthal_invert
from .projection import geo_projection


class StereographicRaw:
    def __call__(self, x: float, y: float) -> Point2D:
        cy = cos(y)
        k = 1 + cos(x) * cy
        return [cy * sin(x) / k, sin(y) / k]

    def invert(self, x: float, y: float) -> Point2D:
        def angle(z: float) -> float:
            return 2 * atan(z)

        return azimuthal_invert(angle)(x, y)


def geo_stereographic() -> Projection:
    """
    The stereographic projection.

    Returns
    -------
    Projection
        Projection object
    """
    return geo_projection(StereographicRaw()).scale(250).set_clip_angle(142)



from bisect import bisect_right
from collections.abc import Callable
from datetime import datetime, timedelta

from ..array import tick_step
from .day import time_day
from .duration import duration_year
from .hour import time_hour
from .interval import TimeInterval
from .millisecond import time_millisecond
from .minute import time_minute
from .month import time_month
from .second import time_second
from .week import time_sunday, time_week
from .year import time_year


class Ticker:
    """
    Tick generator

    Parameters
    ----------
    year : Callable
        Year interval function
    month : Callable
        Month interval function
    week : Callable
        Week interval function
    day : Callable
        Day interval function
    hour : Callable
        Hour interval function
    minute : Callable
        Minute interval function
    """

    def __init__(
        self,
        year: Callable,
        month: Callable,
        week: Callable,
        day: Callable,
        hour: Callable,
        minute: Callable,
    ):
        self.tick_intervals = [
            (time_second, 1, timedelta(seconds=1)),
            (time_second, 5, timedelta(seconds=5)),
            (time_second, 15, timedelta(seconds=15)),
            (time_second, 30, timedelta(seconds=30)),
            (time_minute, 1, timedelta(minutes=1)),
            (time_minute, 5, timedelta(minutes=5)),
            (time_minute, 15, timedelta(minutes=15)),
            (time_minute, 30, timedelta(minutes=30)),
            (time_hour, 1, timedelta(hours=1)),
            (time_hour, 3, timedelta(hours=3)),
            (time_hour, 6, timedelta(hours=6)),
            (time_hour, 12, timedelta(hours=9)),
            (time_day, 1, timedelta(days=1)),
            (time_day, 2, timedelta(days=2)),
            (time_week, 1, timedelta(weeks=1)),
            (time_month, 1, timedelta(days=31)),
            (time_month, 3, timedelta(days=31 * 3)),
            (time_year, 1, timedelta(weeks=52)),
        ]

    def ticks(self, start: datetime, stop: datetime, count: int) -> list[datetime]:
        """
        Returns an array of approximately count dates at regular
        intervals between start and stop (inclusive).

        Parameters
        ----------
        start : datetime
            Start date
        stop : datetime
            Stop date
        count : int
            Count

        Returns
        -------
        list[datetime]
            Array of approximated dates

        Examples
        --------

        If you give *rounded* :code:`datetime` (for example, day times starting
        and finishing at :code:`00:00:00`), the function iterates over days.

        >>> from datetime import datetime
        >>> start = datetime(2011, 1, 1, 0, 0, 0)
        >>> stop = datetime(2011, 1, 5, 0, 0, 0)
        >>> count = 4
        >>> dates = d3.time_ticks(start, stop, count)
        >>> type(dates)
        <class 'list'>
        >>> for d in dates:
        ...     print(d)
        ...
        ...
        2011-01-01 00:00:00
        2011-01-02 00:00:00
        2011-01-03 00:00:00
        2011-01-04 00:00:00
        2011-01-05 00:00:00

        However, if you give *specific* :code:`datetime`, the function rounds
        dates based on the time delta between :code:`stop` and :code:`start`
        and returns :code:`count` dates.

        >>> from datetime import datetime
        >>> start = datetime(2011, 1, 1, 12, 0, 0)
        >>> stop = datetime(2011, 1, 5, 12, 0, 0)
        >>> count = 4
        >>> dates = d3.time_ticks(start, stop, count)
        >>> type(dates)
        <class 'list'>
        >>> for d in dates:
        ...     print(d)
        ...
        ...
        2011-01-02 00:00:00
        2011-01-03 00:00:00
        2011-01-04 00:00:00
        2011-01-05 00:00:00

        As you can see, the closest date to :code:`2011-01-01 12:00:00` is
        :code:`2011-01-02 00:00:00`. The function starts from this date and
        iterates :code:`count` times.
        """
        reverse = stop < start
        if reverse:
            start, stop = stop, start
        interval = (
            count if hasattr(count, "range") else self.tick_interval(start, stop, count)
        )
        ticks = (
            interval.range(start, stop + timedelta(microseconds=1)) if interval else []
        )
        return ticks[::-1] if reverse else ticks

    def tick_interval(
        self, start: datetime, stop: datetime, count: int
    ) -> TimeInterval:
        """
        Returns the time interval that would be used by :func:`d3.time_ticks
        <time_ticks>` given the same arguments

        Parameters
        ----------
        start : datetime
            Start date
        stop : datetime
            Stop date
        count : int
            Count

        Returns
        -------
        TimeInterval
            Time interval chosen used by :func:`d3.time_ticks
            <time_ticks>`
        """
        target = abs(stop - start) / count
        i = bisect_right([step for _, _, step in self.tick_intervals], target)
        if i == len(self.tick_intervals):
            return time_year.every(
                tick_step(
                    start.timestamp() / duration_year,
                    stop.timestamp() / duration_year,
                    count,
                )
                * 1000
            )
        if i == 0:
            return time_millisecond.every(
                max(tick_step(start.timestamp(), stop.timestamp(), count) * 1000, 1)
            )
        if i == len(self.tick_intervals):
            raise ValueError("Too large interval")
        t, step, _ = self.tick_intervals[
            i - 1
            if target / self.tick_intervals[i - 1][2]
            < self.tick_intervals[i][2] / target
            else i
        ]
        return t.every(step)


time = Ticker(time_year, time_month, time_sunday, time_day, time_hour, time_minute)
time_ticks, time_tick_interval = time.ticks, time.tick_interval



from collections.abc import Callable
from itertools import chain
from math import nan

from ...types import Point2D
from ..common import PolygonStream
from .buffer import ClipBuffer
from .line import clip_line
from .rejoin import EPSILON, Intersection, clip_rejoin

clip_max = 1e9
clip_min = -1e9


def clamp(x):
    if x > clip_max:
        return clip_max
    if x < clip_min:
        return clip_min
    return x


class ClipRectangle(PolygonStream):
    def __init__(
        self, x0: float, y0: float, x1: float, y1: float, stream: PolygonStream
    ):
        self._x0 = x0
        self._y0 = y0
        self._x1 = x1
        self._y1 = y1
        self._stream = stream

        self._point = self._point_default

        self._active_stream = stream
        self._buffer_stream = ClipBuffer()
        self._segments = None
        self._polygon = None
        self._fpx = None
        self._fpy = None
        self._fpv = None
        self._ppx = None
        self._ppy = None
        self._ppv = None
        self._ring = None
        self._clean = None

    def _visible(self, x: float, y: float) -> bool:
        return self._x0 <= x <= self._x1 and self._y0 <= y <= self._y1

    def _interpolate(
        self,
        vfrom: float | None,
        vto: float | None,
        direction: float,
        stream: PolygonStream,
    ):
        a = 0
        a1 = 0
        if vfrom is not None:
            a = self._corner(vfrom, direction)
            a1 = self._corner(vto, direction)
        if (
            vfrom is None
            or (a != a1)
            or (self._compare_point(vfrom, vto) < 0) ^ (direction > 0)
        ):
            while True:
                x = self._x0 if a == 0 or a == 3 else self._x1
                y = self._y1 if a > 1 else self._y0
                stream.point(x, y)
                a = (a + direction + 4) % 4
                if a == a1:
                    break
        else:
            stream.point(vto[0], vto[1])

    def _corner(self, p: Point2D, direction: float) -> int:
        if abs(p[0] - self._x0) < EPSILON:
            return 0 if direction > 0 else 3
        elif abs(p[0] - self._x1) < EPSILON:
            return 2 if direction > 0 else 1
        elif abs(p[1] - self._y0) < EPSILON:
            return 1 if direction > 0 else 0
        return 3 if direction > 0 else 2

    def _compare_intersection(self, a: Intersection, b: Intersection) -> float:
        return self._compare_point(a.x, b.x)

    def _compare_point(self, a: Point2D, b: Point2D) -> float:
        ca = self._corner(a, 1)
        cb = self._corner(b, 1)
        if ca != cb:
            return ca - cb
        elif ca == 0:
            return b[1] - a[1]
        elif ca == 1:
            return a[0] - b[0]
        elif ca == 2:
            return a[1] - b[1]
        else:
            return b[0] - a[0]

    def point(self, x: float, y: float):
        return self._point(x, y)

    def _point_default(self, x: float, y: float):
        if self._visible(x, y):
            self._active_stream.point(x, y)

    def polygon_inside(self):
        winding = 0

        for ring in self._polygon:
            point = ring[0]
            b0 = point[0]
            b1 = point[1]
            for j in range(1, len(ring)):
                a0 = b0
                a1 = b1
                point = ring[j]
                b0 = point[0]
                b1 = point[1]
                if a1 <= self._y1:
                    if b1 > self._y1 and (b0 - a0) * (self._y1 - a1) > (b1 - a1) * (
                        self._x0 - a0
                    ):
                        winding += 1
                else:
                    if b1 <= self._y1 and (b0 - a0) * (self._y1 - a1) < (b1 - a1) * (
                        self._x0 - a0
                    ):
                        winding -= 1
        return winding

    def polygon_start(self):
        self._active_stream = self._buffer_stream
        self._segments = []
        self._polygon = []
        self._clean = True

    def polygon_end(self):
        start_inside = self.polygon_inside()
        clean_inside = self._clean and start_inside
        self._segments = list(chain.from_iterable(self._segments))
        visible = len(self._segments)
        if clean_inside or visible:
            self._stream.polygon_start()
            if clean_inside:
                self._stream.line_start()
                self._interpolate(None, None, 1, self._stream)
                self._stream.line_end()
            if visible:
                clip_rejoin(
                    self._segments,
                    self._compare_intersection,
                    start_inside,
                    self._interpolate,
                    self._stream,
                )
            self._stream.polygon_end()
        self._active_stream = self._stream
        self._segments = None
        self._polygon = None
        self._ring = None

    def line_start(self):
        self._point = self._line_point
        if self._polygon is not None:
            self._ring = []
            self._polygon.append(self._ring)
        self._first = True
        self._ppv = False
        self._ppx = nan
        self._ppy = nan

    def line_end(self):
        if self._segments is not None:
            self._line_point(self._fpx, self._fpy)
            if self._fpv and self._ppv:
                self._buffer_stream.rejoin()
            self._segments.append(self._buffer_stream.result())
        self._point = self._point_default
        if self._ppv:
            self._active_stream.line_end()

    def _line_point(self, x: float, y: float):
        v = self._visible(x, y)
        if self._polygon is not None:
            self._ring.append([x, y])
        if self._first:
            self._fpx = x
            self._fpy = y
            self._fpv = v
            self._first = False
            if v:
                self._active_stream.line_start()
                self._active_stream.point(x, y)
        else:
            if v and self._ppv:
                self._active_stream.point(x, y)
            else:
                self._ppx = clamp(self._ppx)
                self._ppy = clamp(self._ppy)
                x = clamp(x)
                y = clamp(y)
                if result := clip_line(
                    self._ppx, self._ppy, x, y, self._x0, self._y0, self._x1, self._y1
                ):
                    x1, y1, x2, y2 = result
                    if not self._ppv:
                        self._active_stream.line_start()
                        self._active_stream.point(x1, y1)
                    self._active_stream.point(x2, y2)
                    if not v:
                        self._active_stream.line_end()
                    self._clean = False
                elif v:
                    self._active_stream.line_start()
                    self._active_stream.point(x, y)
                    self._clean = False
        self._ppx = x
        self._ppy = y
        self._ppv = v

    def __str__(self) -> str:
        return f"ClipRectangle({self._stream}, {[[self._x0, self._y0], [self._x1, self._y1]]})"


class ClipRectangleWrapper:
    def __init__(
        self,
        x0: float,
        y0: float,
        x1: float,
        y1: float,
    ):
        self._x0 = x0
        self._y0 = y0
        self._x1 = x1
        self._y1 = y1

    def __call__(self, stream: PolygonStream):
        return ClipRectangle(self._x0, self._y0, self._x1, self._y1, stream)


def geo_clip_rectangle(
    x0: float,
    y0: float,
    x1: float,
    y1: float,
) -> Callable[[PolygonStream], ClipRectangle]:
    """
    Generates a clipping function which transforms a stream such that
    geometries are bounded by a rectangle of coordinates :math:`[[x_0, y_0],
    [x_1, y_1]]`. Typically used for post-clipping.

    Parameters
    ----------
    x0 : float
        :math:`x_0` coordinate of the rectangle
    y0 : float
        :math:`y_0` coordinate of the rectangle
    x1 : float
        :math:`x_1` coordinate of the rectangle
    y1 : float
        :math:`y_1` coordinate of the rectangle

    Returns
    -------
    Callable[[PolygonStream], ClipRectangle]
        Clipping function
    """

    return ClipRectangleWrapper(x0, y0, x1, y1)



from math import inf

from ...array import argpass
from ...types import Point2D
from ..common import PolygonStream


@argpass
def noop():
    return


class BoundsStream(PolygonStream):
    def __init__(self):
        self._line_start = noop
        self._line_end = noop
        self._polygon_start = noop
        self._polygon_end = noop
        self._x0 = inf
        self._y0 = inf
        self._x1 = -inf
        self._y1 = -inf

    def line_start(self):
        return self._line_start()

    def line_end(self):
        return self._line_end()

    def polygon_start(self):
        return self._polygon_start()

    def polygon_end(self):
        return self._polygon_end()

    def point(self, x: float, y: float):
        return self._bounds_point(x, y)

    def result(self) -> tuple[Point2D, Point2D]:
        bounds = [[self._x0, self._y0], [self._x1, self._y1]]
        self._y0 = self._x0 = inf
        self._x1 = self._y1 = -inf
        return bounds

    def _bounds_point(self, x: float, y: float):
        if x < self._x0:
            self._x0 = x
        if x > self._x1:
            self._x1 = x
        if y < self._y0:
            self._y0 = y
        if y > self._y1:
            self._y1 = y

    def __str__(self):
        return f"BoundsStream({[[self._x0, self._y0], [self._x1, self._y1]]})"



from .albers import geo_albers
from .albers_usa import geo_albers_usa
from .azimuthal_equal_area import geo_azimuthal_equal_area
from .azimuthal_equidistant import geo_azimuthal_equidistant
from .conic_conformal import geo_conic_conformal
from .conic_equal_area import geo_conic_equal_area
from .conic_equidistant import geo_conic_equidistant
from .equal_earth import geo_equal_earth
from .equirectangular import geo_equirectangular
from .gnomonic import geo_gnomonic
from .identity import geo_identity
from .mercator import geo_mercator
from .natural_earth_1 import geo_natural_earth_1
from .orthographic import geo_orthographic
from .projection import geo_projection, geo_projection_mutator
from .stereographic import geo_stereographic
from .transverse_mercator import geo_transverse_mercator

__all__ = [
    "geo_albers",
    "geo_albers_usa",
    "geo_azimuthal_equal_area",
    "geo_azimuthal_equidistant",
    "geo_conic_conformal",
    "geo_conic_equal_area",
    "geo_conic_equidistant",
    "geo_equirectangular",
    "geo_mercator",
    "geo_orthographic",
    "geo_projection",
    "geo_projection_mutator",
    "geo_stereographic",
    "geo_transverse_mercator",
    "geo_gnomonic",
    "geo_natural_earth_1",
    "geo_equal_earth",
    "geo_identity",
]



from math import isnan

from ...types import T
from ..series import Series
from .none import offset_none


def offset_expand(series: list[Series[T]], order: list[int]):
    """
    Applies a zero baseline and normalizes the values for each point such that
    the topline is always one.

    Parameters
    ----------
    series : list[Series[T]]
        List of series
    order : list[int]
        List of ordered indices
    """
    n = len(series)
    if n == 0:
        return
    m = len(series[0])
    for j in range(m):
        y = 0
        for i in range(n):
            x = series[i][j][1]
            y += 0.0 if isnan(x) else x
        if y:
            for i in range(n):
                series[i][j][1] /= y
    offset_none(series, order)



import math
from typing import Any


def threshold_sturges(values: list[float | None], *args: Any) -> int:
    """
    Returns the number of bins according to Sturges' formula; the input values
    must be numbers.

    Parameters
    ----------
    values : list[float | None]
        Input values
    *args: Any
        Unused arguments

    Returns
    -------
    int
       Number of bins according to Sturges' formula
    """
    values = [v for v in values if v is not None]
    if len(values) == 0:
        return 1
    return max(1, math.ceil(math.log(len(values)) / math.log(2)) + 1)



from math import atan, exp, inf, log, pi, tan

from ...types import Point2D
from ..common import Projection, RawProjection
from .mercator import MercatorProjection

half_pi = pi * 0.5


class TransverseMercatorRaw:
    def __call__(self, lambda_: float, phi: float) -> Point2D:
        t = tan((half_pi + phi) / 2)
        return [log(t) if t > 0.0 else -inf, -lambda_]

    def invert(self, x: float, y: float) -> Point2D:
        return [-y, 2 * atan(exp(x)) - half_pi]


class TransverseMercatorProjection(MercatorProjection):
    def __init__(self, project: RawProjection):
        MercatorProjection.__init__(self, project)
        super().rotate([0, 0, 90])
        super().scale(159.155)

    def set_center(self, center: Point2D) -> Projection:
        center = [-center[1], center[0]]
        return super().set_center(center)

    def get_center(self) -> Point2D:
        center = super().get_center()
        return [center[1], -center[0]]

    def rotate(self, angles: tuple[float, float, float]) -> Projection:
        angles = [angles[0], angles[1], angles[2] + 90 if len(angles) > 2 else 90]
        return super().rotate(angles)

    def get_rotation(self) -> tuple[float, float, float]:
        angles = super().get_rotation()
        return [angles[0], angles[1], angles[2] - 90]


def geo_transverse_mercator() -> Projection:
    """
    The transverse spherical Mercator projection. Defines a default
    :func:`Projection.set_clip_extent
    <detroit.geo.common.Projection.set_clip_extent>` such that the world is
    projected to a square, clipped to approximately :math:`\\pm 85°` latitude.

    Returns
    -------
    Projection
        Projection object
    """
    return TransverseMercatorProjection(TransverseMercatorRaw())



from math import asin, cos, sin, sqrt

from ...types import Point2D
from ..common import Projection
from .projection import geo_projection

EPSILON2 = 1e-12
A1 = 1.340264
A2 = -0.081106
A3 = 0.000893
A4 = 0.003796
M = sqrt(3) / 2
iterations = 12


class EqualEarthRaw:
    def __call__(self, lambda_: float, phi: float) -> Point2D:
        l_ = asin(M * sin(phi))
        l2 = l_ * l_
        l6 = l2 * l2 * l2
        return [
            lambda_ * cos(l_) / (M * (A1 + 3 * A2 * l2 + l6 * (7 * A3 + 9 * A4 * l2))),
            l_ * (A1 + A2 * l2 + l6 * (A3 + A4 * l2)),
        ]

    def invert(self, x: float, y: float) -> Point2D:
        l_ = y
        l2 = l_ * l_
        l6 = l2 * l2 * l2
        for i in range(iterations):
            fy = l_ * (A1 + A2 * l2 + l6 * (A3 + A4 * l2)) - y
            fpy = A1 + 3 * A2 * l2 + l6 * (7 * A3 + 9 * A4 * l2)
            delta = fy / fpy
            l_ -= delta
            l2 = l_ * l_
            l6 = l2 * l2 * l2
            if abs(delta) < EPSILON2:
                break
        return [
            M * x * (A1 + 3 * A2 * l2 + l6 * (7 * A3 + 9 * A4 * l2)) / cos(l_),
            asin(sin(l_) / M),
        ]


def geo_equal_earth() -> Projection:
    """
    The Equal Earth projection, an equal-area projection, by Bojan Šavrič et
    al., 2018.

    returns
    -------
    Projection
        Projection object
    """
    return geo_projection(EqualEarthRaw()).scale(177.158)



import math

from ..types import T
from .ticks import tick_increment


def nice(start: T, stop: T, count: int) -> tuple[T, T]:
    """
    Returns a new interval :code:`[nice_start, nice_stop]` covering
    the given interval :code:`[start, stop]` and where :code:`nice_start` and
    :code:`nice_stop` are guaranteed to align with the corresponding
    :code:`tick_step`.

    Parameters
    ----------
    start : T
        Start value
    stop : T
        End value
    count : int
        Count value

    Returns
    -------
    tuple[T, T]
        Aligned interval

    Examples
    --------

    >>> d3.nice(10.2, 20.8, 10)
    [10, 21]
    >>> d3.nice(10.2, 20.8, 1)
    [0, 50]
    """
    prestep = None
    while True:
        step = tick_increment(start, stop, count)
        if step == prestep or step == 0 or not math.isfinite(step):
            return [start, stop]
        elif step > 0:
            start = math.floor(start / step) * step
            stop = math.ceil(stop / step) * step
        elif step < 0:
            start = math.ceil(start * step) / step
            stop = math.floor(stop * step) / step
        prestep = step



from collections.abc import Callable
from math import asin, atan2, cos, fabs, hypot, nan, radians

from ..cartesian import cartesian
from ..common import PolygonStream, Projection
from ..transform import GeoTransformer

EPSILON = 1e-6
max_depth = 16
cos_min_distance = cos(radians(30))


def resample_none(project: Projection) -> GeoTransformer:
    def point(self, x: float, y: float):
        a, b = project(x, y)
        self._stream.point(a, b)

    return GeoTransformer({"point": point})


class Resample(PolygonStream):
    def __init__(self, project: Projection, delta2: float, stream: PolygonStream):
        self._project = project
        self._delta2 = delta2
        self._stream = stream

        self._lambda00 = None
        self._x00 = None
        self._y00 = None
        self._a00 = None
        self._b00 = None
        self._c00 = None
        self._lambda0 = None
        self._x0 = None
        self._y0 = None
        self._a0 = None
        self._b0 = None
        self._c0 = None

        self._point = self._point_default
        self._line_start = self._line_start_default
        self._line_end = self._line_end_default

    def resample_line_to(
        self, x0, y0, lambda0, a0, b0, c0, x1, y1, lambda1, a1, b1, c1, depth
    ):
        dx = x1 - x0
        dy = y1 - y0
        d2 = dx * dx + dy * dy
        if d2 > 4 * self._delta2 and depth:
            depth -= 1
            a = a0 + a1
            b = b0 + b1
            c = c0 + c1
            m = hypot(a, b, c)
            c /= m
            if fabs(fabs(c) - 1.0) < EPSILON or fabs(lambda0 - lambda1) < EPSILON:
                lambda2 = (lambda0 + lambda1) * 0.5
            else:
                lambda2 = atan2(b, a)
            x2, y2 = self._project(lambda2, asin(c))
            dx2 = x2 - x0
            dy2 = y2 - y0
            dz = dy * dx2 - dx * dy2
            if (
                dz * dz / d2 > self._delta2
                or fabs((dx * dx2 + dy * dy2) / d2 - 0.5) > 0.3
                or a0 * a1 + b0 * b1 + c0 * c1 < cos_min_distance
            ):
                a /= m
                b /= m
                self.resample_line_to(
                    x0, y0, lambda0, a0, b0, c0, x2, y2, lambda2, a, b, c, depth
                )
                self._stream.point(x2, y2)
                self.resample_line_to(
                    x2, y2, lambda2, a, b, c, x1, y1, lambda1, a1, b1, c1, depth
                )

    def line_start(self):
        self._line_start()

    def line_end(self):
        self._line_end()

    def point(self, lambda_: float, phi: float):
        return self._point(lambda_, phi)

    def polygon_start(self):
        self._stream.polygon_start()
        self._line_start = self.ring_start

    def polygon_end(self):
        self._stream.polygon_end()
        self._line_start = self._line_start_default

    def _point_default(self, x: float, y: float):
        a, b = self._project(x, y)
        self._stream.point(a, b)

    def _line_start_default(self):
        self._x0 = nan
        self._y0 = nan
        self._point = self._line_point
        self._stream.line_start()

    def _line_point(self, lambda_: float, phi: float):
        a1, b1, c1 = cartesian([lambda_, phi])
        x1, y1 = self._project(lambda_, phi)
        self.resample_line_to(
            self._x0,
            self._y0,
            self._lambda0,
            self._a0,
            self._b0,
            self._c0,
            x1,
            y1,
            lambda_,
            a1,
            b1,
            c1,
            max_depth,
        )
        self._x0 = x1
        self._y0 = y1
        self._lambda0 = lambda_
        self._a0 = a1
        self._b0 = b1
        self._c0 = c1
        self._stream.point(self._x0, self._y0)

    def _line_end_default(self):
        self._point = self._point_default
        self._stream.line_end()

    def ring_start(self):
        self._line_start_default()
        self._point = self.ring_point
        self._line_end = self.ring_end

    def ring_point(self, lambda_: float, phi: float):
        self._lambda00 = lambda_
        self._line_point(lambda_, phi)
        self._x00 = self._x0
        self._y00 = self._y0
        self._a00 = self._a0
        self._b00 = self._b0
        self._c00 = self._c0
        self._point = self._line_point

    def ring_end(self):
        self.resample_line_to(
            self._x0,
            self._y0,
            self._lambda0,
            self._a0,
            self._b0,
            self._c0,
            self._x00,
            self._y00,
            self._lambda00,
            self._a00,
            self._b00,
            self._c00,
            max_depth,
        )
        self._line_end = self._line_end_default
        self._line_end_default()

    def __str__(self):
        return f"Resample({self._stream})"


class ResampleWrapper:
    def __init__(self, project: Projection, delta2: float):
        self._project = project
        self._delta2 = delta2

    def __call__(self, stream: PolygonStream) -> Resample:
        return Resample(self._project, self._delta2, stream)


def resample(
    project: Projection, delta2: float
) -> Callable[[PolygonStream], PolygonStream]:
    if delta2:
        return ResampleWrapper(project, delta2)
    else:
        return resample_none(project)



NAMESPACES = {
    "svg": "http://www.w3.org/2000/svg",
    "xhtml": "http://www.w3.org/1999/xhtml",
    "xlink": "http://www.w3.org/1999/xlink",
    "xml": "http://www.w3.org/XML/1998/namespace",
    "xmlns": "http://www.w3.org/2000/xmlns/",
}


def namespace(name: str):
    prefix = name
    i = prefix.find(":")
    if i >= 0 and (prefix := name[:i]) != "xmlns":
        name = name[i + 1 :]
    return (
        {"space": NAMESPACES[prefix], "local": name} if prefix in NAMESPACES else name
    )



from math import cos, pi, sin

HALF_PI = pi * 0.5


def ease_sin_in(t: float) -> float:
    """
    Sinusoidal easing; returns :code:`sin(t)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_sin_in(0.2)
    0.04894348370484647
    """
    if t == 1:
        return 1
    else:
        return 1 - cos(t * HALF_PI)


def ease_sin_out(t: float) -> float:
    """
    Reverse sinusoidal easing; equivalent to :code:`1 - d3.ease_sin_in(1 - t)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_sin_out(0.2)
    0.3090169943749474
    """
    return sin(t * HALF_PI)


def ease_sin_in_out(t: float) -> float:
    """
    Symmetric sinusoidal easing; scales :func:`d3.ease_sin_in
    <detroit.ease_sin_in>` :math:`\\forall t \\in [0, 0.5]` and
    :func:`d3.ease_sin_out <detroit.ease_sin_out>` :math:`\\forall t \\in
    [0.5, 1]`. Also equivalent to :code:`d3.ease_poly.exponent(3)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_sin_in_out(0.2)
    0.09549150281252627
    """
    return (1 - cos(pi * t)) * 0.5



from math import fsum

from ...array import argpass
from ..common import PolygonStream


@argpass
def noop():
    return


class AreaStream(PolygonStream):
    def __init__(self):
        self._area_sum = []
        self._area_ring_sum = []
        self._point = noop
        self._line_start = noop
        self._line_end = noop

    def line_start(self):
        return self._line_start()

    def line_end(self):
        return self._line_end()

    def polygon_start(self):
        self._line_start = self._area_ring_start
        self._line_end = self._area_ring_end

    def polygon_end(self):
        self._line_start = self._line_end = self._point = noop
        self._area_sum.append(abs(fsum(self._area_ring_sum)))
        self._area_ring_sum = []

    def point(self, x: float, y: float):
        return self._point(x, y)

    def result(self) -> float:
        area = fsum(self._area_sum) * 0.5
        self._area_sum = []
        return area

    def _area_ring_start(self):
        self._point = self._area_point_first

    def _area_point_first(self, x: float, y: float):
        self._point = self._area_point
        self._x00 = x
        self._x0 = x
        self._y00 = y
        self._y0 = y

    def _area_point(self, x: float, y: float):
        self._area_ring_sum.append(self._y0 * x - self._x0 * y)
        self._x0 = x
        self._y0 = y

    def _area_ring_end(self):
        self._area_point(self._x00, self._y00)

    def __str__(self):
        return "AreaStream()"



from ..types import Interval, T


def nice(domain: list[T], interval: Interval) -> list[T]:
    domain = domain.copy()

    i0 = 0
    i1 = len(domain) - 1
    x0 = domain[i0]
    x1 = domain[i1]
    t = None

    if x1 < x0:
        t = i0
        i0 = i1
        i1 = t
        t = x0
        x0 = x1
        x1 = t

    domain[i0] = interval.floor(x0)
    domain[i1] = interval.ceil(x1)
    return domain



from ..colors import colors
from ..ramp import ramp

scheme = [
    "fee8c8fdbb84e34a33",
    "fef0d9fdcc8afc8d59d7301f",
    "fef0d9fdcc8afc8d59e34a33b30000",
    "fef0d9fdd49efdbb84fc8d59e34a33b30000",
    "fef0d9fdd49efdbb84fc8d59ef6548d7301f990000",
    "fff7ecfee8c8fdd49efdbb84fc8d59ef6548d7301f990000",
    "fff7ecfee8c8fdd49efdbb84fc8d59ef6548d7301fb300007f0000",
]
SCHEME_ORRD = list(map(colors, scheme))
interpolate_orrd = ramp(SCHEME_ORRD)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "efedf5bcbddc756bb1",
    "f2f0f7cbc9e29e9ac86a51a3",
    "f2f0f7cbc9e29e9ac8756bb154278f",
    "f2f0f7dadaebbcbddc9e9ac8756bb154278f",
    "f2f0f7dadaebbcbddc9e9ac8807dba6a51a34a1486",
    "fcfbfdefedf5dadaebbcbddc9e9ac8807dba6a51a34a1486",
    "fcfbfdefedf5dadaebbcbddc9e9ac8807dba6a51a354278f3f007d",
]
SCHEME_PURPLES = list(map(colors, scheme))
interpolate_purples = ramp(SCHEME_PURPLES)



from __future__ import annotations

import math
from bisect import bisect
from collections.abc import Callable
from statistics import quantiles
from typing import overload

from .continuous import identity
from .init import init_interpolator


class SequentialQuantile:
    def __init__(self):
        self._domain = []
        self._interpolator = identity

    def __call__(self, x: int | float) -> int | float:
        """
        Given a value from the domain, returns the corresponding value from the range.

        Parameters
        ----------
        x : int | float
            Input value

        Returns
        -------
        int | float
            Output value
        """
        if x is not None and not (isinstance(x, float) and math.isnan(x)):
            return self._interpolator(
                (bisect(self.get_domain(), x, 1) - 1) / (len(self.get_domain()) - 1)
            )

    def set_domain(self, domain: list[int | float]) -> SequentialQuantile:
        """
        Sets the scale's domain to the specified array of numbers

        Parameters
        ----------
        domain : list[int | float]
            Domain

        Returns
        -------
        SequentialQuantile
            Itself
        """
        self._domain.clear()
        for d in domain:
            if d is not None and not (isinstance(d, float) and math.isnan(d)):
                self._domain.append(d)
        self._domain = sorted(self._domain)
        return self

    def get_domain(self) -> list[int | float]:
        return self._domain.copy()

    def set_interpolator(self, interpolator: Callable) -> SequentialQuantile:
        """
        Sets the scale's interpolator function

        Parameters
        ----------
        interpolator : Callable
            Interpolator function

        Returns
        -------
        SequentialQuantile
            Itself
        """
        self._interpolator = interpolator
        return self

    def get_interpolator(self) -> Callable:
        return self._interpolator

    def get_range(self) -> list[float]:
        """
        Returns an array of interpolated values from domain values
        using interpolator function.

        Returns
        -------
        list[float]
            Interpolated values
        """
        return [
            self._interpolator(i / (len(self.get_domain()) - 1))
            for i in range(len(self.get_domain()))
        ]

    def quantiles(self, n: int) -> list[float]:
        """
        Returns an array of :math:`n + 1` quantiles.

        Parameters
        ----------
        n : int
            :math:`n` value

        Returns
        -------
        list[float]
            Array of :math:`n + 1` quantiles
        """
        return (
            [self.get_domain()[0]]
            + quantiles(self.get_domain(), n=n, method="inclusive")
            + [self.get_domain()[-1]]
        )

    def copy(self):
        return SequentialQuantile().set_domain(self.get_domain())

    def __str__(self) -> str:
        name = self.__class__.__name__
        attrbs = ["domain", "range"]
        attrbs = (f"{a}={getattr(self, f'get_{a}')()}" for a in attrbs)
        attrbs = ", ".join(attrbs)
        return f"{name}({attrbs})"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        addr = id(self)
        return f"<{name} at {hex(addr)}>"


@overload
def scale_sequential_quantile() -> SequentialQuantile: ...


@overload
def scale_sequential_quantile(domain: list[int | float]) -> SequentialQuantile: ...


def scale_sequential_quantile(*args):
    """
    Returns a new sequential scale with a p-quantile
    transform, analogous to a quantile scale.

    Parameters
    ----------
    domain : list[int | float]
        Array of numbers

    Returns
    -------
    SequentialQuantile
        Scale object

    Examples
    --------

    >>> scale = d3.scale_sequential_quantile().set_domain([3, 6, 7, 8, 8, 10, 13, 15, 16, 20])
    >>> scale = scale.set_interpolator(d3.interpolate_blues)
    >>> for x in range(21):
    ...     print(x, scale(x))
    ...
    ...
    0 rgb(247, 251, 255)
    1 rgb(247, 251, 255)
    2 rgb(247, 251, 255)
    3 rgb(247, 251, 255)
    4 rgb(247, 251, 255)
    5 rgb(247, 251, 255)
    6 rgb(225, 237, 248)
    7 rgb(202, 222, 240)
    8 rgb(130, 186, 219)
    9 rgb(130, 186, 219)
    10 rgb(89, 161, 207)
    11 rgb(89, 161, 207)
    12 rgb(89, 161, 207)
    13 rgb(55, 135, 192)
    14 rgb(55, 135, 192)
    15 rgb(28, 106, 175)
    16 rgb(11, 77, 148)
    17 rgb(11, 77, 148)
    18 rgb(11, 77, 148)
    19 rgb(11, 77, 148)
    20 rgb(8, 48, 107)
    >>> d3.interpolate_blues(0)
    'rgb(247, 251, 255)'
    >>> d3.interpolate_blues(1)
    'rgb(8, 48, 107)'
    """
    scale = SequentialQuantile()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    return init_interpolator(scale)



from __future__ import annotations

import math
from typing import Any, Generic

from ..types import T
from .linear import LinearBase


class Identity(LinearBase, Generic[T]):
    """
    Build a new identity scale with the specified range (and by extension, domain).
    """

    def __init__(self, domain=None):
        self._domain = list(domain) if domain is not None else [0, 1]
        self._unknown = None

    def __call__(self, x: T) -> T:
        """
        Returns same value if valid type

        Parameters
        ----------
        x : T
            Input value

        Returns
        -------
        T
            Input value if valid type
        """
        unvalid_type = x is None or (isinstance(x, float) and math.isnan(x))
        return self._unknown if unvalid_type else x

    def invert(self, x: T) -> T:
        """
        Returns same value if valid type

        Parameters
        ----------
        x : T
            Input value

        Returns
        -------
        T
            Input value if valid type
        """
        return self(x)

    def set_domain(self, domain: list[T]) -> Identity:
        """
        Sets the scale’s domain

        Parameters
        ----------
        domain : list[T]
            Domain

        Returns
        -------
        Identity
            Itself
        """
        self._domain = list(domain)
        return self

    def get_domain(self):
        return self._domain

    def set_range(self, range_vals: list[T]) -> Identity:
        """
        Sets the scale’s range

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        Identity
            Itself
        """
        return self.set_domain(range_vals)

    def get_range(self):
        return self._domain

    def set_unknown(self, unknown: Any) -> Identity:
        """
        Sets the output value of the scale for undefined
        or NaN input values.

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        Identity
            Itself
        """
        self._unknown = unknown
        return self

    def get_unknown(self):
        return self._unknown

    def copy(self):
        return Identity(self._domain).set_unknown(self._unknown)

    def __str__(self) -> str:
        name = self.__class__.__name__
        attrbs = ["domain", "range"]
        attrbs = (f"{a}={getattr(self, f'get_{a}')()}" for a in attrbs)
        attrbs = ", ".join(attrbs)
        return f"{name}({attrbs})"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        addr = id(self)
        return f"<{name} at {hex(addr)}>"


def scale_identity(values: list[T] | None = None) -> Identity[T]:
    """
    Build a new identity scale with the specified range
    (and by extension, domain).

    Parameters
    ----------
    values : list[T] | None
        Values to set

    Returns
    -------
    Identity
        Scale object

    Examples
    --------

    >>> scale = d3.scale_identity()
    >>> scale(3)
    3
    >>> scale("a")
    'a'
    """
    return Identity(domain=values)



from math import asin, inf, sqrt

from ...types import Point2D
from ..common import Projection
from .azimuthal import azimuthal_invert, azimuthal_raw
from .projection import geo_projection


class AzimuthalEqualAreaRaw:
    def __call__(self, lambda_: float, phi: float) -> Point2D:
        def scale(cxcy: float) -> float:
            return sqrt(2 / (1 + cxcy)) if cxcy != -1 else inf

        return azimuthal_raw(scale)(lambda_, phi)

    def invert(self, x: float, y: float) -> Point2D:
        def angle(z: float) -> float:
            return 2 * asin(z / 2)

        return azimuthal_invert(angle)(x, y)


def geo_azimuthal_equal_area() -> Projection:
    """
    The azimuthal equal-area projection.

    Returns
    -------
    Projection
        Projection object
    """
    return (
        geo_projection(AzimuthalEqualAreaRaw()).scale(124.75).set_clip_angle(180 - 1e-3)
    )



import math

from ...selection import Selection
from .common import Curve, fdiv, isvaluable, sign


class HermiteInterpolation:
    def _slope3(self, x2, y2):
        h0 = self._x1 - self._x0
        h1 = x2 - self._x1
        s0 = fdiv(self._y1 - self._y0, h0)
        s0 = -s0 if h1 < 0 else s0
        s1 = fdiv(y2 - self._y1, h1)
        s1 = -s1 if h0 < 0 else s1
        p = fdiv(s0 * h1 + s1 * h0, h0 + h1)
        r = (sign(s0) + sign(s1)) * min(abs(s0), abs(s1), 0.5 * abs(p))
        return 0.0 if math.isnan(r) or math.isinf(r) else r

    def _slope2(self, t):
        h = self._x1 - self._x0
        return (3 * (self._y1 - self._y0) / h - t) * 0.5 if h else t

    def _hermite_point(self, t0, t1):
        x0 = self._x0
        y0 = self._y0
        x1 = self._x1
        y1 = self._y1
        dx = (x1 - x0) / 3
        self._context.bezier_curve_to(
            x0 + dx, y0 + dx * t0, x1 - dx, y1 - dx * t1, x1, y1
        )


class Monotone(Curve, HermiteInterpolation):
    def __init__(self, context):
        self._context = context
        self._line = math.nan
        self._x0 = math.nan
        self._y0 = math.nan
        self._x1 = math.nan
        self._y1 = math.nan
        self._t0 = math.nan

    def area_start(self):
        self._line = 0

    def area_end(self):
        self._line = math.nan

    def line_start(self):
        self._x0 = math.nan
        self._y0 = math.nan
        self._x1 = math.nan
        self._y1 = math.nan
        self._t0 = math.nan
        self._point = 0

    def line_end(self):
        if self._point == 2:
            self._context.line_to(self._x1, self._y1)
        elif self._point == 3:
            self._hermite_point(self._t0, self._slope2(self._t0))

        if isvaluable(self._line) or (self._line != 0 and self._point == 1):
            self._context.close_path()

        self._line = 1 - self._line

    def point(self, x, y):
        t1 = math.nan
        if x == self._x1 and y == self._y1:
            return

        if self._point == 0:
            self._point = 1
            if isvaluable(self._line):
                self._context.line_to(x, y)
            else:
                self._context.move_to(x, y)
        elif self._point == 1:
            self._point = 2
        elif self._point == 2:
            self._point = 3
            t1 = self._slope3(x, y)
            self._hermite_point(self._slope2(t1), t1)
        else:
            t1 = self._slope3(x, y)
            self._hermite_point(self._t0, t1)

        self._x0 = self._x1
        self._x1 = x
        self._y0 = self._y1
        self._y1 = y
        self._t0 = t1


class ReflectContext:
    def __init__(self, context):
        self._context = context

    def move_to(self, x, y):
        self._context.move_to(y, x)

    def line_to(self, x, y):
        self._context.line_to(y, x)

    def close_path(self):
        self._context.close_path()

    def bezier_curve_to(self, x1, y1, x2, y2, x, y):
        self._context.bezier_curve_to(y1, x1, y2, x2, y, x)


class MonotoneX(Monotone):
    def __init__(self, context):
        Monotone.__init__(self, context)


class MonotoneY(Monotone):
    def __init__(self, context):
        Monotone.__init__(self, ReflectContext(context))

    def point(self, x, y):
        super().point(y, x)


def curve_monotone_x(context: Selection) -> Curve:
    """
    Produces a cubic spline that preserves monotonicity in y, assuming
    monotonicity in x, as proposed by Steffen in A simple method for monotonic
    interpolation in one dimension: "a smooth curve with continuous first-order
    derivatives that passes through any given set of data points without
    spurious oscillations. Local extrema can occur only at grid points where
    they are given by the data, but not in between two adjacent grid points."

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_monotone_x)
    >>> line(points)
    'M2,2C3.333,4.333,4.667,6.667,6,8C7.333,9.333,8.667,10,10,10C10.667,10,11.333,10,12,10C12.667,10,13.333,4,14,4C16,4,18,4,20,4C21.333,4,22.667,8,24,8C25.667,8,27.333,6.944,29,6C30,5.433,31,4,32,4C33,4,34,5,35,5C36,5,37,3.500,38,2'

    **Result**

    .. image:: ../../figures/light_curve_monotone_x.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_monotone_x.svg
       :align: center
       :class: only-dark
    """
    return MonotoneX(context)


def curve_monotone_y(context: Selection) -> Curve:
    """
    Produces a cubic spline that preserves monotonicity in x, assuming
    monotonicity in y, as proposed by Steffen in A simple method for monotonic
    interpolation in one dimension: "a smooth curve with continuous first-order
    derivatives that passes through any given set of data points without
    spurious oscillations. Local extrema can occur only at grid points where
    they are given by the data, but not in between two adjacent grid points."

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_monotone_y)
    >>> line(points)
    'M2,2C2.667,4,3.333,6,6,8C6.889,8.667,7.333,9.333,10,10C10,10,12,10,12,10C13.333,8,12.667,6,14,4C14,4,20,4,20,4C22.667,5.333,26.667,6.667,24,8C25.333,7.333,30.333,6.667,29,6C27.667,5.333,30,4.667,32,4C31,4.333,35.667,4.667,35,5C37,4,37.500,3,38,2'

    **Result**

    .. image:: ../../figures/light_curve_monotone_y.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_monotone_y.svg
       :align: center
       :class: only-dark
    """
    return MonotoneY(context)



from collections.abc import Callable
from io import StringIO
from math import isnan, nan

from ..common import PolygonStream


def asint(x: float) -> int | float:
    return int(x) if x.is_integer() else x


def round_zero(x: float) -> int:
    return int(round(x))


def round_digits(digits: int) -> Callable[[float], float]:
    def local_round(x: float) -> float:
        return asint(float(round(x, digits)))

    return local_round


def identity(x: float) -> float:
    return asint(x)


class PathString(PolygonStream):
    def __init__(self, digits: int | None = None):
        digits = 15 if digits is None else digits
        if isnan(digits) or digits < 0:
            raise ValueError(f"Invalid digits: {digits}")

        self._digits = digits
        self._round = identity
        if self._digits == 0:
            self._round = round_zero
        elif self._digits < 15:
            self._round = round_digits(self._digits)

        self._cache_append = None
        self._cache_radius = None
        self._cache_circle = None

        self._radius = 4.5
        self._string = StringIO(newline="")
        self._line = nan
        self._point = nan

    def point_radius(self, radius: float):
        self._radius = radius
        return self

    def polygon_start(self):
        self._line = 0

    def polygon_end(self):
        self._line = nan

    def line_start(self):
        self._point = 0

    def line_end(self):
        if self._line == 0:
            self._string.write("Z")
        self._point = nan

    def point(self, x: float, y: float):
        match self._point:
            case 0:
                x = self._round(x)
                y = self._round(y)
                self._string.write(f"M{x},{y}")
                self._point = 1
            case 1:
                x = self._round(x)
                y = self._round(y)
                self._string.write(f"L{x},{y}")
            case _:
                x = self._round(x)
                y = self._round(y)
                self._string.write(f"M{x},{y}")
                if self._radius != self._cache_radius:
                    r = self._round(self._radius)
                    r2 = self._round(2 * self._radius)
                    s = self._string
                    self._string = StringIO(newline="")
                    self._string.write(
                        f"m0,{r}a{r},{r} 0 1,1 0,{-r2}a{r},{r} 0 1,1 0,{r2}z"
                    )
                    self._cache_radius = r
                    self._cache_circle = self._string
                    self._string = s
                self._string.write(self._cache_circle.getvalue())

    def result(self) -> str | None:
        result = self._string.getvalue()
        self._string = StringIO(newline="")
        return result if len(result) else None

    def __str__(self) -> str:
        return "PathString()"



from math import asin, atan2, cos, degrees, radians, sin, sqrt

from ..types import Point2D


def haversin(x: float) -> float:
    x = sin(x * 0.5)
    return x * x


class Interpolate:
    def __init__(self, a: Point2D, b: Point2D):
        self._x0 = radians(a[0])
        self._y0 = radians(a[1])
        self._x1 = radians(b[0])
        self._y1 = radians(b[1])
        self._cy0 = cos(self._y0)
        self._sy0 = sin(self._y0)
        self._cy1 = cos(self._y1)
        self._sy1 = sin(self._y1)
        self._kx0 = self._cy0 * cos(self._x0)
        self._ky0 = self._cy0 * sin(self._x0)
        self._kx1 = self._cy1 * cos(self._x1)
        self._ky1 = self._cy1 * sin(self._x1)
        self._d = 2 * asin(
            sqrt(
                haversin(self._y1 - self._y0)
                + self._cy0 * self._cy1 * haversin(self._x1 - self._x0)
            )
        )
        self._k = sin(self._d)
        self._interpolate = (
            self._interpolate_non_null if self._d else self._interpolate_null
        )

    def __call__(self, t: float) -> Point2D:
        return self._interpolate(t)

    def _interpolate_non_null(self, t: float) -> Point2D:
        t *= self._d
        B = sin(t) / self._k
        A = sin(self._d - t) / self._k
        x = A * self._kx0 + B * self._kx1
        y = A * self._ky0 + B * self._ky1
        z = A * self._sy0 + B * self._sy1
        return [degrees(atan2(y, x)), degrees(atan2(z, sqrt(x * x + y * y)))]

    def _interpolate_null(self, t: float) -> Point2D:
        return [degrees(self._x0), degrees(self._y0)]

    def get_distance(self) -> float:
        return self._d


def geo_interpolate(a: Point2D, b: Point2D) -> Interpolate:
    """
    Returns an interpolator function given two points a and b. Each point must
    be specified as a two-element array :code:`[longitude, latitude]` in
    degrees. The returned interpolator function takes a single argument t,
    where t is a number ranging from 0 to 1; a value of 0 returns the point a,
    while a value of 1 returns the point b. Intermediate values interpolate
    from a to b along the great arc that passes through both a and b. If a and
    b are antipodes, an arbitrary great arc is chosen.

    Parameters
    ----------
    a : Point2D
        2D point
    b : Point2D
        2D point

    Returns
    -------
    Interpolate
        Interpolate function
    """
    return Interpolate(a, b)



from math import sqrt

SQRT_3 = sqrt(3)


def symbol_triangle2(context, size):
    s = sqrt(size) * 0.6824
    t = s / 2
    u = (s * SQRT_3) / 2
    context.move_to(0, -s)
    context.line_to(u, t)
    context.line_to(-u, t)
    context.close_path()



from __future__ import annotations

from collections.abc import Callable, Iterator
from functools import cmp_to_key
from math import pi
from typing import Any

TAU = 2 * pi


class ChordValue:
    """
    Chord subgroup

    Attributes
    ----------
    index : int
        The node index i
    start_angle : float
        The start angle in radians
    end_angle : float
        The end angle in radians
    value : float
        The flow value matrix[i][j]
    """

    __slots__ = ("index", "start_angle", "end_angle", "value")

    def __init__(self, index: int, start_angle: float, end_angle: float, value: float):
        self.index = index
        self.start_angle = start_angle
        self.end_angle = end_angle
        self.value = value

    def get(self, key: str, default: Any | None = None) -> int | float | None:
        match key:
            case "index":
                return self.index
            case "start_angle":
                return self.start_angle
            case "end_angle":
                return self.end_angle
            case "value":
                return self.value
            case _:
                return default

    def __getitem__(self, key: str) -> int | float:
        value = self.get(key)
        if value is None:
            raise KeyError(f"'ChordValue' object has no attribute {key!r}")
        return value

    def __str__(self) -> str:
        return f"ChordValue({self.index}, {self.start_angle}, {self.end_angle}, {self.value})"

    def __repr__(self) -> str:
        return str(self)


class ChordItem:
    """
    Chord object containing source and target subgroups

    Attributes
    ----------
    source : ChordValue | None
        The source subgroup
    target : ChordValue | None
        The target subgroup
    """

    __slots__ = ("source", "target")

    def __init__(
        self, source: ChordValue | None = None, target: ChordValue | None = None
    ):
        self.source = source
        self.target = target

    def get(self, key: str, default: Any | None = None) -> int | float | None:
        match key:
            case "source":
                return self.source
            case "target":
                return self.target
            case _:
                return default

    def __getitem__(self, key: str) -> int | float:
        value = self.get(key)
        if value is None:
            raise KeyError(f"'ChordItem' object has no attribute {key!r}")
        return value

    def __str__(self) -> str:
        return f"ChordItem({self.source}, {self.target})"

    def __repr__(self) -> str:
        return str(self)


class Chords:
    __slots__ = ("_data", "groups")

    def __init__(self, data: list[ChordItem], groups: list[ChordValue]):
        self._data = data
        self.groups = groups

    def __len__(self):
        return len(self._data)

    def __iter__(self) -> Iterator[ChordItem]:
        return iter(self._data)

    def __getitem__(self, index: int) -> ChordItem:
        return self._data[index]

    def sort(self, *, key=None, reverse: bool = False):
        self._data.sort(key=key, reverse=reverse)


def compare_value(
    compare: Callable[[float, float], float],
) -> Callable[[ChordValue, ChordValue], float]:
    def local_compare(a: ChordValue, b: ChordValue):
        return compare(
            a.source.value + a.source.value,
            b.source.value + b.source.value,
        )


class Chord:
    """
    Chord layout

    Parameters
    ----------
    directed : bool
        :code:`True` for directed flow
    transpose : bool
        :code:`True` to transpose the specified matrix
    """

    def __init__(self, directed: bool, transpose: bool):
        self._directed = directed
        self._transpose = transpose
        self._pad_angle = 0
        self._sort_groups = None
        self._sort_subgroups = None
        self._sort_chords = None

    def __call__(self, matrix: list[list[float]]) -> Chords:
        """
        Computes the chord layout for the specified square matrix of size
        :math:`n \\times n`, where the matrix represents the directed flow
        amongst a network (a complete digraph) of n nodes.

        The return value of chord(matrix) is an array of chords, where each
        chord represents the combined bidirectional flow between two nodes i
        and j (where i may be equal to j) and is an object with the following
        properties:

        * :code:`source` - the source subgroup
        * :code:`target` - the target subgroup

        Each source and target subgroup is also an object with the following
        properties:

        * :code:`index` - the node index i
        * :code:`start_angle` - the start angle in radians
        * :code:`end_angle` - the end angle in radians
        * :code:`value` - the flow value matrix[i][j]

        The chords are typically passed to ribbon to display the network
        relationships.

        The returned array includes only chord objects for which the value
        :code:`matrix[i][j]` or :code:`matrix[j][i]` is non-zero. Furthermore,
        the returned array only contains unique chords: a given chord ij
        represents the bidirectional flow from i to j and from j to i, and does
        not contain a duplicate chord ji; i and j are chosen such that the
        chord's source always represents the larger of :code:`matrix[i][j]` and
        :code:`matrix[j][i]`.

        The chords array also defines a secondary array of length n,
        chords.groups, where each group represents the combined outflow for
        node i, corresponding to the elements :code:`matrix[i][0 ... n - 1]`,
        and is an object with the following properties:

        * :code:`index` - the node index i
        * :code:`start_angle` - the start angle in radians
        * :code:`end_angle` - the end angle in radians
        * :code:`value` - the total outgoing flow value for node i

        The groups are typically passed to arc to produce a donut chart around
        the circumference of the chord layout.

        Parameters
        ----------
        matrix : list[list[float]]
            Square matrix of size :math:`n \\times n`

        Returns
        -------
        Chords
            Chords Object which behaves like a list and has
            :code:`Chords.groups` attribute. Groups are typically passed to arc
            to produce a donut chart around the circumference of the chord
            layout.
        """
        n = len(matrix)
        group_sums = [None] * n
        group_index = list(range(n))
        chords = [None] * (n * n)
        groups = [None] * n
        k = 0

        matrix = (
            [matrix[i % n][i // n] for i in range(n * n)]
            if self._transpose
            else [matrix[i // n][i % n] for i in range(n * n)]
        )

        for i in range(n):
            x = 0
            for j in range(n):
                x += matrix[i * n + j] + self._directed * matrix[j * n + i]
            group_sums[i] = x
            k += x
        k = max(0, TAU - self._pad_angle * n) / k
        dx = self._pad_angle if k else TAU / n

        x = 0
        if self._sort_groups:
            group_index.sort(
                key=cmp_to_key(
                    lambda a, b: self._sort_groups(group_sums[a], group_sums[b])
                )
            )
        for i in group_index:
            x0 = x
            if self._directed:
                subgroup_index = list(
                    filter(
                        lambda j: (matrix[-j * n + i] if j < 0 else matrix[i * n + j]),
                        range(-n + 1, n),
                    )
                )
                if self._sort_subgroups:
                    subgroup_index.sort(
                        key=cmp_to_key(
                            lambda a, b: self._sort_subgroups(
                                (-matrix[-a * n + i] if a < 0 else matrix[i * n + a]),
                                (-matrix[-b * n + i] if b < 0 else matrix[i * n + b]),
                            )
                        )
                    )
                for j in subgroup_index:
                    if j < 0:
                        chord = chords[-j * n + i]
                        if chord is None:
                            chord = chords[-j * n + i] = ChordItem()
                        x_ = x
                        x += matrix[-j * n + i] * k
                        chord.target = ChordValue(i, x_, x, matrix[-j * n + i])
                    else:
                        chord = chords[i * n + j]
                        if chord is None:
                            chord = chords[i * n + j] = ChordItem()
                        x_ = x
                        x += matrix[i * n + j] * k
                        chord.source = ChordValue(i, x_, x, matrix[i * n + j])
                groups[i] = ChordValue(i, x0, x, group_sums[i])
            else:
                subgroup_index = list(
                    filter(lambda j: matrix[i * n + j] or matrix[j * n + i], range(n))
                )
                if self._sort_subgroups:
                    subgroup_index.sort(
                        key=cmp_to_key(
                            lambda a, b: self._sort_subgroups(
                                matrix[i * n + a], matrix[i * n + b]
                            )
                        )
                    )
                for j in subgroup_index:
                    if i < j:
                        chord = chords[i * n + j]
                        if chord is None:
                            chord = chords[i * n + j] = ChordItem()
                        x_ = x
                        x += matrix[i * n + j] * k
                        chord.source = ChordValue(i, x_, x, matrix[i * n + j])
                    else:
                        chord = chords[j * n + i]
                        if chord is None:
                            chord = chords[j * n + i] = ChordItem()
                        x_ = x
                        x += matrix[i * n + j] * k
                        chord.target = ChordValue(i, x_, x, matrix[i * n + j])
                        if i == j:
                            chord.source = chord.target
                    if (
                        chord.source
                        and chord.target
                        and chord.source.value < chord.target.value
                    ):
                        chord.source, chord.target = chord.target, chord.source
                groups[i] = ChordValue(i, x0, x, group_sums[i])
            x += dx

        idx = 0
        while idx < len(chords):
            if chords[idx] is None:
                chords.pop(idx)
            else:
                idx += 1
        chords = Chords(chords, groups)
        if self._sort_chords:
            chords.sort(key=cmp_to_key(self._sort_chords))
        return chords

    def set_pad_angle(self, pad_angle: float) -> Chord:
        """
        Sets the pad angle between adjacent groups to the specified number in
        radians and returns this chord layout.

        Parameters
        ----------
        pad_angle : float
            Pad angle value

        Returns
        -------
        Chord
            Itself
        """
        self._pad_angle = max(0, pad_angle)
        return self

    def set_sort_groups(
        self, sort_groups: Callable[[float, float], float] | None = None
    ) -> Chord:
        """
        Sets the group comparator to the specified function or :code:`None` and
        returns this chord layout.

        Parameters
        ----------
        sort_groups : Callable[[float, float], float] | None
            Sort group function

        Returns
        -------
        Chord
            Itself
        """
        self._sort_groups = sort_groups
        return self

    def set_sort_subgroups(
        self, sort_subgroups: Callable[[float, float], float] | None = None
    ) -> Chord:
        """
        Sets the subgroup comparator to the specified function or :code:`None`
        and returns this chord layout.

        Parameters
        ----------
        sort_subgroups : Callable[[float, float], float] | None
            Sort subgroups function

        Returns
        -------
        Chord
            Itself
        """
        self._sort_subgroups = sort_subgroups
        return self

    def set_sort_chords(
        self, compare: Callable[[float, float], float] | None = None
    ) -> Chord:
        """
        Sets the chord comparator to the specified function or :code:`None` and
        returns this chord layout.

        Parameters
        ----------
        compare : Callable[[float, float], float] | None
            Compare function

        Returns
        -------
        Chord
            Itself
        """
        if compare is None:
            self._sort_chords = None
        else:
            self._sort_chords = compare_value(compare)
        return self

    def get_pad_angle(self) -> float:
        return self._pad_angle

    def get_sort_groups(self) -> Callable[[float, float], float]:
        return self._sort_groups

    def get_sort_subgroups(self) -> Callable[[float, float], float]:
        return self._sort_subgroups

    def get_sort_chords(self) -> Callable[[ChordItem, ChordItem], float]:
        return self._sort_chords


def chord() -> Chord:
    """
    Constructs a new chord layout with the default settings.

    Returns
    -------
    Chord
        Chord object
    """
    return Chord(False, False)


def chord_transpose() -> Chord:
    """
    A transposed chord layout. Useful to highlight outgoing (rather than
    incoming) flows.

    Returns
    -------
    Chord
        Chord object
    """
    return Chord(False, True)


def chord_directed() -> Chord:
    """
    A chord layout for unidirectional flows. The chord from i to j is generated
    from the value in :code:`matrix[i][j]` only.

    Returns
    -------
    Chord
        Chord object
    """
    return Chord(True, False)



from __future__ import annotations

import math
from collections.abc import Callable
from datetime import datetime
from typing import Any, Generic, overload

from ..interpolate import interpolate, interpolate_round
from ..types import Number, T
from .continuous import identity
from .init import init_interpolator
from .linear import LinearBase
from .log import LogBase, logp, powp, reflect, transform_log, transform_logn
from .pow import transform_pow, transform_sqrt
from .symlog import transform_symlog


class Sequential(Generic[T]):
    """
    Sequential transformation

    Parameters
    ----------
    t : Callable[[Number], T]
        Transform function
    """

    def __init__(self, t: Callable[[Number], T]):
        self._x0 = 0
        self._x1 = 1
        self._transform = t
        self._t0 = t(self._x0)
        self._t1 = t(self._x1)
        self._k10 = 0 if self._t0 == self._t1 else 1 / (self._t1 - self._t0)
        self._interpolator = identity
        self._clamp = False
        self._unknown = None

    def __call__(self, x: Number) -> T:
        """
        Given a value from the domain, returns the corresponding value from the range.

        Parameters
        ----------
        x : Number
            Input value

        Returns
        -------
        T
            Corresponding value from the range
        """
        if x is None or (isinstance(x, float) and math.isnan(x)):
            return self._unknown
        if self._k10 == 0:
            x = 0.5
        else:
            x = (self._transform(x) - self._t0) * self._k10
            if self._clamp:
                x = max(0, min(1, x))
        return self._interpolator(x)

    def set_domain(self, domain: list[Number]) -> Sequential:
        """
        Sets the scale's domain to the specified array of numbers

        Parameters
        ----------
        domain : list[Number]
            Domain

        Returns
        -------
        Sequential
            Itself
        """
        self._x0, self._x1 = map(float, list(domain)[:2])
        self._t0 = self._transform(self._x0)
        self._t1 = self._transform(self._x1)
        self._k10 = 0 if self._t0 == self._t1 else 1 / (self._t1 - self._t0)
        return self

    def get_domain(self) -> list[Number]:
        return [self._x0, self._x1]

    def set_clamp(self, clamp: bool) -> Sequential:
        """
        Enables or disables clamping accordingly.

        Parameters
        ----------
        clamp : bool
            Clamp value

        Returns
        -------
        Sequential
            Itself
        """
        self._clamp = bool(clamp)
        return self

    def get_clamp(self) -> bool:
        return self._clamp

    def set_interpolator(self, interpolator: Callable[[float], float]) -> Sequential:
        """
        Sets the scale’s interpolator to the specified function.

        Parameters
        ----------
        interpolator: Callable[[float], float]
            Interpolator function

        Returns
        -------
        Sequential
            Itself
        """
        self._interpolator = interpolator
        return self

    def get_interpolator(self) -> Callable[[float], float]:
        return self._interpolator

    def set_range(self, range_vals: list[T]) -> Sequential:
        """
        The given two-element array is converted
        to an interpolator function using interpolate

        Parameters
        ----------
        range_vals : list[T]
            Two values

        Returns
        -------
        Sequential
            Itself
        """
        self._r0, self._r1 = list(range_vals)[:2]
        self._interpolator = interpolate(self._r0, self._r1)
        return self

    def get_range(self) -> list[T]:
        return [self._interpolator(0), self._interpolator(1)]

    def set_range_round(self, range_vals: list[T]) -> Sequential:
        """
        Sets the scale's range to the specified array of values
        and sets scale's interpolator to :code:`interpolate_round`.

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        Sequential
            Itself
        """
        self._r0, self._r1 = range_vals
        self._interpolator = interpolate_round(self._r0, self._r1)
        return self

    def get_range_round(self) -> list[T]:
        return [self._interpolator(0), self._interpolator(1)]

    def set_unknown(self, unknown: Any) -> Sequential:
        """
        Sets the output value of the scale for undefined
        or NaN input values.

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        Sequential
            Itself
        """
        self._unknown = unknown
        return self

    def get_unknown(self) -> Any:
        return self._unknown

    def __str__(self) -> str:
        name = self.__class__.__name__
        attrbs = ["domain", "range"]
        attrbs = (f"{a}={getattr(self, f'get_{a}')()}" for a in attrbs)
        attrbs = ", ".join(attrbs)
        return f"{name}({attrbs})"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        addr = id(self)
        return f"<{name} at {hex(addr)}>"


def copy(source, target):
    return (
        target.set_domain(source.get_domain())
        .set_interpolator(source.get_interpolator())
        .set_clamp(source.get_clamp())
        .set_unknown(source.get_unknown())
    )


class SequentialLinear(Sequential[float], LinearBase):
    """
    Linear sequential transformation
    """

    def __init__(self):
        Sequential.__init__(self, identity)

    def copy(self):
        return copy(self, SequentialLinear())


class SequentialLog(Sequential[float], LogBase):
    """
    Log sequential transformation
    """

    def __init__(self):
        Sequential.__init__(self, identity)
        LogBase.__init__(self)
        self._transform = transform_log
        self.set_domain([1, 10])

    def _rescale(self):
        self._logs = logp(self._base)
        self._pows = powp(self._base)
        d = self.get_domain()[0]
        if isinstance(d, datetime):
            d = d.timestamp()
        if d < 0:
            self._logs = reflect(self._logs)
            self._pows = reflect(self._pows)
            self._transform = transform_logn
        else:
            self._transform = transform_log
        return self

    def set_domain(self, domain: list[Number]) -> SequentialLog:
        """
        Sets the scale's domain to the specified array of numbers

        Parameters
        ----------
        domain : list[Number]
            Domain

        Returns
        -------
        SequentialLog
            Itself
        """
        self._x0, self._x1 = map(float, list(domain)[:2])
        self._rescale()
        super().set_domain(domain)
        return self

    def copy(self):
        return copy(self, SequentialLog()).base(self.base)


class SequentialSymlog(Sequential[float]):
    """
    Symlog sequential transformation

    Parameters
    ----------
    c : Number
        Symlog constant value
    """

    def __init__(self, c: Number = 1):
        self._c = c
        super().__init__(transform_symlog(self._c))

    def set_constant(self, c: Number) -> SequentialSymlog:
        """
        Sets the symlog constant to the specified number and returns this scale.

        Parameters
        ----------
        c : Number
            Constant value

        Returns
        -------
        SequentialSymlog
            Itself
        """
        self._c = float(c)
        self._transform = transform_symlog(self._c)
        return self

    def copy(self):
        return copy(self, SequentialSymlog()).set_constant(self.constant)


class SequentialPow(Sequential[float], LinearBase):
    """
    Power sequential transformation

    Parameters
    ----------
    t : Callable[[Number], T]
        Transform function
    """

    def __init__(self, t: Callable[[Number], float] = identity):
        super().__init__(t)
        self._exponent = 1

    def _rescale(self) -> SequentialPow:
        if self._exponent == 1:
            self._transform = identity
            return self
        elif self._exponent == 0.5:
            self._transform = transform_sqrt
            return self
        else:
            self._transform = transform_pow(self._exponent)
            return self

    def set_exponent(self, exponent: Number) -> SequentialPow:
        """
        Sets the scale's exponent value.

        Parameters
        ----------
        exponent : Number
            Exponent value

        Returns
        -------
        SequentialPow
            Itself
        """
        self._exponent = float(exponent)
        return self._rescale()

    def get_exponent(self):
        return self._exponent

    def copy(self):
        return copy(self, SequentialPow()).set_exponent(self.exponent)


@overload
def scale_sequential() -> SequentialLinear: ...


@overload
def scale_sequential(interpolator: Callable[[float], float]) -> SequentialLinear: ...


@overload
def scale_sequential(
    domain: list[Number], interpolator: Callable[[float], float]
) -> SequentialLinear: ...


def scale_sequential(*args):
    """
    Builds a new sequential scale with the specified
    domain and interpolator function or array.

    Parameters
    ----------
    domain : list[Number]
        Domain
    interpolator: Callable[[float], float]
        Interpolator

    Returns
    -------
    SequentialLinear
        Sequential object

    Examples
    --------

    >>> scale = d3.scale_sequential([0, 100], d3.interpolate_blues)
    >>> for x in range(11):
    ...     x = 10 * x
    ...     print(x, scale(x))
    ...
    ...
    0 rgb(247, 251, 255)
    10 rgb(227, 238, 249)
    20 rgb(207, 225, 242)
    30 rgb(181, 212, 233)
    40 rgb(147, 195, 223)
    50 rgb(109, 174, 213)
    60 rgb(75, 151, 201)
    70 rgb(47, 126, 188)
    80 rgb(24, 100, 170)
    90 rgb(10, 74, 144)
    100 rgb(8, 48, 107)
    >>> d3.interpolate_blues(0)
    'rgb(247, 251, 255)'
    >>> d3.interpolate_blues(1)
    'rgb(8, 48, 107)'
    """
    scale = SequentialLinear()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_sequential_log() -> SequentialLog: ...


@overload
def scale_sequential_log(interpolator: Callable[[float], float]) -> SequentialLog: ...


@overload
def scale_sequential_log(
    domain: list[Number], interpolator: Callable[[float], float]
) -> SequentialLog: ...


def scale_sequential_log(*args):
    """
    Builds a new sequential scale with a logarithmic
    transform, analogous to a log scale.

    Parameters
    ----------
    domain : list[Number]
        Domain
    interpolator: Callable[[float], float]
        Interpolator

    Returns
    -------
    SequentialLog
        Sequential object

    Examples
    --------

    >>> scale = d3.scale_sequential_log([1, 100], d3.interpolate_blues)
    >>> steps = 10
    >>> for x in range(steps + 1):
    ...     x = 2 * x / steps
    ...     x = 10 ** x
    ...     print(x, scale(x))
    ...
    ...
    1.0 rgb(247, 251, 255)
    1.5848931924611136 rgb(227, 238, 249)
    2.51188643150958 rgb(207, 225, 242)
    3.9810717055349722 rgb(181, 212, 233)
    6.309573444801933 rgb(147, 195, 223)
    10.0 rgb(109, 174, 213)
    15.848931924611133 rgb(75, 151, 201)
    25.118864315095795 rgb(47, 126, 188)
    39.810717055349734 rgb(24, 100, 170)
    63.09573444801933 rgb(10, 74, 144)
    100.0 rgb(8, 48, 107)
    >>> d3.interpolate_blues(0)
    'rgb(247, 251, 255)'
    >>> d3.interpolate_blues(1)
    'rgb(8, 48, 107)'
    """
    scale = SequentialLog()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_sequential_symlog() -> SequentialSymlog: ...


@overload
def scale_sequential_symlog(
    interpolator: Callable[[float], float],
) -> SequentialSymlog: ...


@overload
def scale_sequential_symlog(
    domain: list[Number], interpolator: Callable[[float], float]
) -> SequentialSymlog: ...


def scale_sequential_symlog(*args):
    """
    Builds a new sequential scale with a symmetric
    logarithmic transform, analogous to a symlog scale.

    Parameters
    ----------
    domain : list[Number]
        Domain
    interpolator: Callable[[float], float]
        Interpolator

    Returns
    -------
    SequentialSymlog
        Sequential object

    Examples
    --------

    >>> scale = d3.scale_sequential_symlog([1, 100], d3.interpolate_blues)
    >>> scale = scale.set_constant(2)
    >>> steps = 10
    >>> for x in range(steps + 1):
    ...     x = 2 * x / steps
    ...     x = 10 ** x
    ...     print(x, scale(x))
    ...
    ...
    1.0 rgb(255, 255, 255)
    1.5848931924611136 rgb(253, 255, 255)
    2.51188643150958 rgb(241, 247, 253)
    3.9810717055349722 rgb(227, 238, 248)
    6.309573444801933 rgb(210, 227, 243)
    10.0 rgb(187, 215, 235)
    15.848931924611133 rgb(154, 199, 225)
    25.118864315095795 rgb(113, 177, 214)
    39.810717055349734 rgb(75, 152, 201)
    63.09573444801933 rgb(44, 123, 186)
    100.0 rgb(19, 94, 165)
    >>> d3.interpolate_blues(0)
    'rgb(247, 251, 255)'
    >>> d3.interpolate_blues(1)
    'rgb(8, 48, 107)'
    """
    scale = SequentialSymlog()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_sequential_pow() -> SequentialPow: ...


@overload
def scale_sequential_pow(interpolator: Callable[[float], float]) -> SequentialPow: ...


@overload
def scale_sequential_pow(
    domain: list[Number], interpolator: Callable[[float], float]
) -> SequentialPow: ...


def scale_sequential_pow(*args):
    """
    Builds a new sequential scale with an exponential
    transform, analogous to a power scale.


    Parameters
    ----------
    domain : list[Number]
        Domain
    interpolator: Callable[[float], float]
        Interpolator

    Returns
    -------
    SequentialPow
        Sequential object

    Examples
    --------

    >>> scale = d3.scale_sequential_pow([0, 100], d3.interpolate_blues)
    >>> scale = scale.set_exponent(2)
    >>> steps = 10
    >>> for x in range(steps + 1):
    ...     x = 10 * x / steps
    ...     print(x, scale(x))
    ...
    ...
    0.0 rgb(247, 251, 255)
    1.0 rgb(245, 250, 254)
    2.0 rgb(239, 246, 252)
    3.0 rgb(229, 239, 249)
    4.0 rgb(215, 231, 245)
    5.0 rgb(195, 219, 238)
    6.0 rgb(162, 203, 227)
    7.0 rgb(112, 176, 214)
    8.0 rgb(63, 141, 196)
    9.0 rgb(22, 98, 168)
    10.0 rgb(8, 48, 107)
    >>> d3.interpolate_blues(0)
    'rgb(247, 251, 255)'
    >>> d3.interpolate_blues(1)
    'rgb(8, 48, 107)'
    """
    scale = SequentialPow()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_sequential_sqrt() -> SequentialPow: ...


@overload
def scale_sequential_sqrt(interpolator: Callable[[float], float]) -> SequentialPow: ...


@overload
def scale_sequential_sqrt(
    domain: list[Number], interpolator: Callable[[float], float]
) -> SequentialPow: ...


def scale_sequential_sqrt(*args):
    """
    Builds a new sequential scale with a square-root
    transform, analogous to a sqrt scale


    Parameters
    ----------
    domain : list[Number]
        Domain
    interpolator: Callable[[float], float]
        Interpolator

    Returns
    -------
    SequentialSqrt
        Sequential object

    Examples
    --------

    >>> scale = d3.scale_sequential_sqrt([0, 100], d3.interpolate_blues)
    >>> steps = 10
    >>> for x in range(steps + 1):
    ...     x = 100**2 * x / steps
    ...     print(x, scale(x))
    ...
    ...
    0.0 rgb(247, 251, 255)
    1000.0 rgb(176, 210, 232)
    2000.0 rgb(129, 186, 219)
    3000.0 rgb(92, 163, 208)
    4000.0 rgb(65, 143, 197)
    5000.0 rgb(45, 124, 186)
    6000.0 rgb(29, 107, 175)
    7000.0 rgb(17, 91, 162)
    8000.0 rgb(11, 76, 146)
    9000.0 rgb(8, 62, 127)
    10000.0 rgb(8, 48, 107)
    >>> d3.interpolate_blues(0)
    'rgb(247, 251, 255)'
    >>> d3.interpolate_blues(1)
    'rgb(8, 48, 107)'
    """
    return scale_sequential_pow(*args).set_exponent(0.5)



from __future__ import annotations

import math
from bisect import bisect
from typing import Any, Generic, overload

from ..types import T
from .init import init_range


class ScaleThreshold(Generic[T]):
    """
    Threshold scales are similar to quantize scales, except they allow
    you to map arbitrary subsets of the domain to discrete values in
    the range. The input domain is still continuous, and divided into
    slices based on a set of threshold values. See this choropleth for
    an example.
    """

    def __init__(self):
        self._domain = [0.5]
        self._range_vals = [0, 1]
        self._unknown = None
        self._n = 1

    def __call__(self, x: int | float | None = None) -> T:
        """
        Given a value in the input domain, returns the corresponding
        value in the output range.

        Parameters
        ----------
        x : int | float | None
            Input value

        Returns
        -------
        T
            Output value
        """
        if x is not None and not (isinstance(x, float) and math.isnan(x)):
            return self._range_vals[bisect(self._domain, x, 0, self._n)]
        else:
            return self._unknown

    def set_domain(self, domain: list[int | float]) -> ScaleThreshold:
        """
        Sets the scale’s domain to the specified array of values.

        Parameters
        ----------
        domain : list[int | float]
            Domain

        Returns
        -------
        ScaleThreshold
            Itself
        """
        self._domain = list(domain)
        self._n = min(len(self._domain), len(self._range_vals) - 1)
        return self

    def get_domain(self) -> list[int | float]:
        return self._domain.copy()

    def set_range(self, range_vals: list[T]) -> ScaleThreshold:
        """
        Sets the scale’s range to the specified array of values.

        Parameters
        ----------
        range_vals : list[T]
            Range

        Returns
        -------
        ScaleThreshold
            Itself
        """
        self._range_vals = list(range_vals)
        self._n = min(len(self._domain), len(self._range_vals) - 1)
        return self

    def get_range(self) -> list[T]:
        return self._range_vals.copy()

    def invert_extent(self, y: T) -> list[int | float | None]:
        """
        Returns the extent of values in the domain :math:`[x_0, x_1]`
        for the corresponding value in the range, representing the
        inverse mapping from range to domain.

        Parameters
        ----------
        y : T
            Input value

        Returns
        -------
        list[int | float | None]
            :math:`[x_0, x_1]` values
        """
        if y not in self._range_vals:
            return [None, None]
        i = self._range_vals.index(y)
        if i == 0:
            return [None, self._domain[i]]
        if i == len(self._domain):
            return [self._domain[i - 1], None]
        return [self._domain[i - 1], self._domain[i]]

    def set_unknown(self, unknown: Any) -> ScaleThreshold:
        """
        Sets the scale's unknown value.

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        ScaleThreshold
            Itself
        """
        self._unknown = unknown
        return self

    def get_unknown(self) -> Any:
        return self._unknown

    def copy(self):
        return (
            ScaleThreshold()
            .set_domain(self.get_domain())
            .set_range(self.get_range())
            .set_unknown(self.get_unknown())
        )

    def __str__(self) -> str:
        name = self.__class__.__name__
        attrbs = ["domain", "range"]
        attrbs = (f"{a}={getattr(self, f'get_{a}')()}" for a in attrbs)
        attrbs = ", ".join(attrbs)
        return f"{name}({attrbs})"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        addr = id(self)
        return f"<{name} at {hex(addr)}>"


@overload
def scale_threshold() -> ScaleThreshold[T]: ...


@overload
def scale_threshold(range_vals: list[T]) -> ScaleThreshold[T]: ...


@overload
def scale_threshold(
    domain: list[int | float], range_vals: list[T]
) -> ScaleThreshold[T]: ...


def scale_threshold(*args):
    """
    Builds a new threshold scale with the specified domain and range.

    Parameters
    ----------
    domain : list[int | float]
        Array of numbers
    range_vals : list[int | float]
        Array of values

    Returns
    -------
    ScaleThreshold[T]
        Scale object

    Examples
    --------

    >>> scale = d3.scale_threshold([0, 1], ["red", "white", "blue"])
    >>> for x in range(steps + 1):
    ...     x = -1 + 2 * x / steps
    ...     print(x, scale(x))
    ...
    ...
    -1.0 red
    -0.8 red
    -0.6 red
    -0.4 red
    -0.19999999999999996 red
    0.0 white
    0.19999999999999996 white
    0.3999999999999999 white
    0.6000000000000001 white
    0.8 white
    1.0 blue
    """
    scale = ScaleThreshold()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "edf8b17fcdbb2c7fb8",
    "ffffcca1dab441b6c4225ea8",
    "ffffcca1dab441b6c42c7fb8253494",
    "ffffccc7e9b47fcdbb41b6c42c7fb8253494",
    "ffffccc7e9b47fcdbb41b6c41d91c0225ea80c2c84",
    "ffffd9edf8b1c7e9b47fcdbb41b6c41d91c0225ea80c2c84",
    "ffffd9edf8b1c7e9b47fcdbb41b6c41d91c0225ea8253494081d58",
]
SCHEME_YLGNBU = list(map(colors, scheme))
interpolate_ylgnbu = ramp(SCHEME_YLGNBU)



import math

from ...selection import Selection
from .common import Curve, isvaluable


class LinearCurve(Curve):
    def __init__(self, context):
        self._context = context
        self._line = math.nan

    def area_start(self):
        self._line = 0

    def area_end(self):
        self._line = math.nan

    def line_start(self):
        self._point = 0

    def line_end(self):
        if isvaluable(self._line) or (self._line != 0 and self._point == 1):
            self._context.close_path()
        self._line = 1 - self._line

    def point(self, x, y):
        if self._point == 0:
            self._point = 1
            if isvaluable(self._line):
                self._context.line_to(x, y)
            else:
                self._context.move_to(x, y)
        elif self._point == 1:
            self._point = 2
            self._context.line_to(x, y)
        else:
            self._context.line_to(x, y)


def curve_linear(context: Selection) -> Curve:
    """
    Produces a polyline through the specified points.

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_linear)
    >>> line(points)
    'M2,2L6,8L10,10L12,10L14,4L20,4L24,8L29,6L32,4L35,5L38,2'

    **Result**

    .. image:: ../../figures/light_curve_linear.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_linear.svg
       :align: center
       :class: only-dark
    """
    return LinearCurve(context)
